'''Calibration constants and dataclasses'''

from dataclasses import dataclass
from enum import Enum

class NumOfCh(Enum):
    '''
    Number of channels for each module
    '''
    DAC = 0x8
    ADC = 0x8
    RTD = 0x1
    TC = 0x1

@dataclass
class CalibParam:
    '''
    Calibration parameters Gains and offsets
    '''
    gain: float = None
    offset: float = None

@dataclass
class ReferenceV:
    '''
    Reference voltage of each module
    '''
    dac: float = None
    adc: float = None
    adc_gnd: float = None



"""unit tests for spi.py module"""

import pytest
from edgepi.peripherals.spi import SpiDevice


@pytest.mark.parametrize(
    "dev_id, bus_num, result",
    [
        (0, 6, "/dev/spidev6.0"),
        (1, 6, "/dev/spidev6.1"),
        (2, 6, "/dev/spidev6.2"),
        (3, 6, "/dev/spidev6.3"),
    ],
)
def test_check_range(mocker, dev_id, bus_num, result):
    mocker.patch("edgepi.peripherals.spi.SPI")
    spi = SpiDevice(bus_num, dev_id)
    assert spi.devpath == result



"""
Module for interacting with the EdgePi Thermocouple via SPI.
"""


import logging
import time
from enum import Enum

from bitstring import Bits
from edgepi.peripherals.spi import SpiDevice
from edgepi.tc.tc_commands import code_to_temp, TempCode, tempcode_to_opcode, TempType
from edgepi.tc.tc_constants import (
    NUM_WRITE_REGS,
    AvgMode,
    CJHighMask,
    CJLowMask,
    CJMode,
    ConvMode,
    DecBits4,
    DecBits6,
    FaultMode,
    NoiseFilterMode,
    OpenCircuitMode,
    OpenMask,
    OvuvMask,
    TCAddresses,
    TCHighMask,
    TCLowMask,
    TCOps,
    TCType,
    VoltageMode,
)
from edgepi.tc.tc_faults import map_fault_status
from edgepi.reg_helper.reg_helper import OpCode, apply_opcodes
from edgepi.utilities.utilities import filter_dict
from edgepi.tc.tc_conv_time import calc_conv_time

_logger = logging.getLogger(__name__)


class EdgePiTC(SpiDevice):
    """
    A class used to represent the EdgePi Thermocouple as an SPI device.
    """

    # default MAX31856 register values for writeable registers
    default_reg_values = {
        TCAddresses.CR0_W.value: 0x00,
        TCAddresses.CR1_W.value: 0x03,
        TCAddresses.MASK_W.value: 0xFF,
        TCAddresses.CJHF_W.value: 0x7F,
        TCAddresses.CJLF_W.value: 0xC0,
        TCAddresses.LTHFTH_W.value: 0x7F,
        TCAddresses.LTHFTL_W.value: 0xFF,
        TCAddresses.LTLFTH_W.value: 0x80,
        TCAddresses.LTLFTL_W.value: 0x00,
        TCAddresses.CJTO_W.value: 0x00,
        TCAddresses.CJTH_W.value: 0x00,
        TCAddresses.CJTL_W.value: 0x00,
    }

    def __init__(self):
        super().__init__(bus_num=6, dev_id=2)

    def read_temperatures(self):
        """Use to read cold junction and linearized thermocouple temperature measurements"""
        temp_bytes = self.__read_registers(TCAddresses.CJTH_R.value, 5)
        return code_to_temp(temp_bytes)

    def single_sample(self, safe_delay: bool = True):
        """Conduct a single sampling event. Returns measured temperature in degrees Celsius.

        Args:
            safe_delay (bool): manual sampling requires a time delay between triggering
                            the sampling and the measured values being available for reading.
                            Set to True to enable the maximum time delay, or False to enable
                            a shorter time delay. Note, using the shorter time delay offers
                            a higher chance of reading temperature values before the new
                            measurement is available.
        Returns:
            a tuple containing temperatures for cold junction
            and linearized thermocouple temperature
        """
        cr0_value = self.__read_register(TCAddresses.CR0_R.value)
        cr1_value = self.__read_register(TCAddresses.CR1_R.value)
        command = cr0_value[1] | TCOps.SINGLE_SHOT.value.op_code
        self.__write_to_register(TCAddresses.CR0_W.value, command)

        # compute time delay between register write and update
        conv_time = calc_conv_time(cr0_value[1], cr1_value[1], safe_delay)
        time.sleep(conv_time / 1000)

        # read cold junction and linearized TC temperatures
        temp_codes = self.read_temperatures()

        _logger.debug(f"single sample codes: {temp_codes}")

        return temp_codes

    def read_faults(self, filter_at_fault=True) -> list:
        """Read information about thermocouple fault status.

        Args:
            filter_at_fault (bool): set to True to return only Faults that are currently asserting

        Returns:
            a dictionary mapping each thermocouple fault type to a Fault object holding
            information about its current status.
            See :obj:`tc.tc_faults.Fault` for details about the Fault class.
        """
        # read in values from fault status register and fault mask register
        faults = self.__read_register(TCAddresses.SR_R.value)
        fault_bits = Bits(uint=faults[1], length=8)
        masks = self.__read_register(TCAddresses.MASK_R.value)
        fault_masks = Bits(uint=masks[1], length=8)

        fault_msgs = map_fault_status(fault_bits, fault_masks)
        _logger.info(f"read_faults:\n{fault_msgs}")

        # filter out normal status Fault objects
        if filter_at_fault:
            fault_msgs = {
                fault_type: fault
                for (fault_type, fault) in fault_msgs.items()
                if fault.at_fault is True
            }

        return fault_msgs

    def clear_faults(self):
        """
        When thermocouple is in Interrupt Fault Mode, clears all bits in Fault Status Register,
        and deasserts FAULT pin output. Note that the FAULT output and the fault bit may
        reassert immediately if the fault persists. If thermocouple is in Comparator Fault Mode,
        this will have no effect on the thermocouple.
        """
        cr0 = self.__read_register(TCAddresses.CR0_R.value)
        command = cr0[1] | TCOps.CLEAR_FAULTS.value.op_code
        self.__write_to_register(TCAddresses.CR0_W.value, command)

    def overwrite_cold_junction_temp(self, cj_temp: int, cj_temp_decimals: DecBits6):
        """
        Write temperature values to the cold-junction sensor.
        Cold-junction sensing must be disabled (using set_config method)
        in order for values to be written to the cold-junction sensor.

        Args:
            cj_temp (int): the integer value of the temperature
                            to be written to the cold-junction sensor

            cj_temp_decimals (DecBits6): the decimal value of the temperature
                                        to be written to the cold-junction sensor

        Raises:
            ColdJunctionOverwriteError: if value is written to cold-junction temperature
                                    registers while cold-junction sensing is not disabled.
        """
        self.set_config(cj_temp=cj_temp, cj_temp_decimals=cj_temp_decimals)

    def reset_registers(self):
        """
        Resets register values to factory default values. Please refer to MAX31856
        datasheet or this module's documentation for these values. Note this will
        not reset the CJTH and CJTL registers, as these require cold-junction
        sensing to be disabled in order to update the values.
        """
        for addx, value in self.default_reg_values.items():
            self.__write_to_register(addx, value)

    def __read_register(self, reg_addx):
        """Reads the value of a single register.

        Args:
            reg_addx (TCAddress.Enum.value): the register's address

        Returns:
            a list containing two entries:
            new_data[0] = register address, new_data[1] = register value
        """
        data = [reg_addx] + [0xFF]
        _logger.debug(f"__read_register: addx = {reg_addx} => data before xfer = {data}")
        new_data = self.transfer(data)
        _logger.debug(f"__read_register: addx = {reg_addx} => data after xfer = {new_data}")
        return new_data

    def __read_registers(self, start_addx: int = 0, regs_to_read: int = 16):
        """read a variable number of registers sequentially

        Args:
            start_addx (TCAddress.Enum.value): address of the register to begin the read at.
            regs_to_read (int): number of registers to read, including starting register.

        Returns:
            a list containing register values starting from start_addx. Note, first entry
            is the start address: register values begin from the second entry.
        """
        data = [start_addx] + [0xFF] * regs_to_read
        _logger.debug(f"__read_registers: shifting in data => {data}")
        new_data = self.transfer(data)
        _logger.debug(f"__read_registers: shifted out data => {new_data}")
        return new_data

    def __write_to_register(self, reg_addx: int, value: int):
        """write a value to a register.

        Args:
            reg_addx (TCAddress.Enum.value): address of the register to write the value to.

            value (int): a values to be written to the register.
        """
        data = [reg_addx] + [value]
        _logger.debug(f"__write_to_registers: shifting in data => {data}")
        new_data = self.transfer(data)
        _logger.debug(f"__write_to_registers: shifted out data => {new_data}")

    def __read_registers_to_map(self):
        """
        Builds a map of write register address to corresponding read register value.
        Note, each register has a read and write address, but only the read address
        contains the register's value. Write addresses are only for writing.

        Returns:
            a dictionary containing (write_reg_address: register_value) entries
            for each writeable register.
        """
        reg_map = {}
        num_regs = NUM_WRITE_REGS
        read_regs_offset = 0x80
        start_addx = TCAddresses.CR0_W.value
        # read values from __read_registers, but log values to corresponding write registers
        reg_values = self.__read_registers(start_addx - read_regs_offset)
        for addx_offset in range(num_regs):
            reg_map[start_addx + addx_offset] = reg_values[
                addx_offset + 1
            ]  # reg_values[0] is start_addx
        _logger.debug(f"__read_registers_to_map => {reg_map}")
        return reg_map

    def __update_registers_from_dict(self, reg_values: dict):
        """Applies updated register values contained in a dictionary of register values

        Args:
            reg_values (dict): a dictionary containing { register_address: entry } pairs, where
                                entry is a dictionary holding 'value' and 'is_changed' keys.
        """
        for reg_addx, entry in reg_values.items():
            if entry["is_changed"]:
                updated_value = entry["value"]
                self.__write_to_register(reg_addx, updated_value)
                _logger.debug(
                    f"""register value at address ({hex(reg_addx)})
                     has been updated to ({hex(updated_value)})"""
                )

    def __get_tc_type(self):
        """Returns the currently configured thermocouple type"""
        cr1 = Bits(uint=self.__read_register(TCAddresses.CR1_R.value)[1], length=8)
        tc_bits = cr1[-4:].uint
        for enum in TCType:
            if enum.value.op_code == tc_bits:
                return enum
        return None

    def __get_cj_status(self):
        "Returns the current cold-junction sensing status (on/off)"
        cr0 = Bits(uint=self.__read_register(TCAddresses.CR0_R.value)[1], length=8)
        return cr0[4]

    def __process_temperature_settings(
        self,
        cj_high_threshold: int,
        cj_low_threshold: int,
        lt_high_threshold: int,
        lt_high_threshold_decimals: DecBits4,
        lt_low_threshold: int,
        lt_low_threshold_decimals: DecBits4,
        cj_offset: int,
        cj_offset_decimals: DecBits4,
        cj_temp: int,
        cj_temp_decimals: DecBits6,
        tc_type: TCType,
        ops_list: list,
    ):
        """generates OpCodes for temperature settings and appends to external OpCodes list"""
        tempcodes = [
            TempCode(
                cj_high_threshold,
                DecBits4.P0,
                7,
                0,
                0,
                TCAddresses.CJHF_W.value,
                TempType.COLD_JUNCTION,
            ),
            TempCode(
                cj_low_threshold,
                DecBits4.P0,
                7,
                0,
                0,
                TCAddresses.CJLF_W.value,
                TempType.COLD_JUNCTION,
            ),
            TempCode(
                lt_high_threshold,
                lt_high_threshold_decimals,
                11,
                4,
                0,
                TCAddresses.LTHFTH_W.value,
                TempType.THERMOCOUPLE,
            ),
            TempCode(
                lt_low_threshold,
                lt_low_threshold_decimals,
                11,
                4,
                0,
                TCAddresses.LTLFTH_W.value,
                TempType.THERMOCOUPLE,
            ),
            TempCode(
                cj_offset,
                cj_offset_decimals,
                3,
                4,
                0,
                TCAddresses.CJTO_W.value,
                TempType.COLD_JUNCTION_OFFSET,
            ),
            TempCode(
                cj_temp, cj_temp_decimals, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION
            ),
        ]

        # in case the user updates thermocouple type as well
        tc_type = tc_type if tc_type is not None else self.__get_tc_type()
        cj_status = self.__get_cj_status()

        for tempcode in tempcodes:
            ops_list += tempcode_to_opcode(tempcode, tc_type, cj_status)
        _logger.debug(f"set_config: ops_list:\n\n{ops_list}\n\n")

    def set_config(
        self,
        conversion_mode: ConvMode = None,
        open_circuit_mode: OpenCircuitMode = None,
        cold_junction_mode: CJMode = None,
        fault_mode: FaultMode = None,
        noise_filter_mode: NoiseFilterMode = None,
        average_mode: AvgMode = None,
        tc_type: TCType = None,
        voltage_mode: VoltageMode = None,
        cj_high_mask: CJHighMask = None,
        cj_low_mask: CJLowMask = None,
        tc_high_mask: TCHighMask = None,
        tc_low_mask: TCLowMask = None,
        ovuv_mask: OvuvMask = None,
        open_mask: OpenMask = None,
        cj_high_threshold: int = None,
        cj_low_threshold: int = None,
        lt_high_threshold: int = None,
        lt_high_threshold_decimals: DecBits4 = None,
        lt_low_threshold: int = None,
        lt_low_threshold_decimals: DecBits4 = None,
        cj_offset: int = None,
        cj_offset_decimals: DecBits4 = None,
        cj_temp: int = None,
        cj_temp_decimals: DecBits6 = None,
    ):
        """
        A collective thermocouple settings update method.
        Use this method to configure thermocouple settings.
        This method allows you to configure settings either individually,
        or collectively (more than one at a time).

        Args:
            conversion_mode (ConvMode): enable manual or automatic sampling

            oc_fault_mode (OpenCircuitMode): set open circuit fault detection mode

            cold_junction_mode (CJMode): enable or disable cold junction sensor

            fault_mode (FaultMode): set fault reading mode

            noise_filter_mode (NoiseFilterMode): set which noise frequency to reject

            average_mode (AvgMode): number of samples to average per temperature measurement

            tc_type (TCType): set thermocouple type

            voltage_mode (VoltageMode): set input voltage range

            cj_high_mask (CJHighMask): mask CJHIGH fault from asserting through the FAULT pin

            cj_low_mask (CJLowMask): mask CJLOW fault from asserting through the FAULT pin

            tc_high_mask (TCHighMask): mask TCHIGH fault from asserting through the FAULT pin

            tc_low_mask (TCLowMask): mask TCLOW fault from asserting through the FAULT pin

            ovuv_mask (OvuvMask): mask OVUV fault from asserting through the FAULT pin

            open_mask (OpenMask): mask OPEN fault from asserting through the FAULT pin

            cj_high_threshold (int): set cold junction temperature upper threshold.
                If cold junction temperature rises above this limit, the FAULT output will assert

            cj_low_threshold (int): set cold junction temperature lower threshold.
                If cold junction temperature falls below this limit, the FAULT output will assert

            lt_high_threshold (int): set thermocouple hot junction temperature upper threshold.
                If thermocouple hot junction temperature rises above this limit,
                the FAULT output will assert

                lt_high_threshold_decimals (DecBits4): set thermocouple hot junction temperature
                                                        upper threshold decimal value.

            lt_low_threshold (int): set thermocouple hot junction temperature lower threshold.
                If thermocouple hot junction temperature falls below this limit,
                the FAULT output will assert

                lt_low_threshold_decimals (DecBits4): set thermocouple hot junction temperature
                                                    lower threshold decimal value.

            cj_offset (int): set cold junction temperature offset.

                cj_offset_decimals (DecBits4): set cold junction temperature offset decimal value.

                cj_temp (int): write values to cold-junction sensor.
                                Only use when cold-junction is disabled.

                cj_temp_decimals (DecBits6): set decimal value for cj_temp
        """
        # pylint: disable=unused-argument

        # filter out self from args
        args_dict = filter_dict(locals(), "self", None)
        _logger.debug(f"set_config: args dict:\n\n {args_dict}\n\n")

        # extract non-temperature setting opcodes from Enums
        ops_list = [
            entry.value
            for entry in args_dict.values()
            if issubclass(entry.__class__, Enum) and isinstance(entry.value, OpCode)
        ]

        self.__process_temperature_settings(
            cj_high_threshold,
            cj_low_threshold,
            lt_high_threshold,
            lt_high_threshold_decimals,
            lt_low_threshold,
            lt_low_threshold_decimals,
            cj_offset,
            cj_offset_decimals,
            cj_temp,
            cj_temp_decimals,
            tc_type,
            ops_list,
        )

        # read value of every write register into dict, starting from CR0_W.
        # Tuples are (write register addx : register_value) pairs.
        reg_values = self.__read_registers_to_map()
        _logger.debug(f"set_config: register values before updates:\n\n{reg_values}\n\n")

        # updated register values
        apply_opcodes(reg_values, ops_list)
        _logger.debug(f"set_config: register values after updates:\n\n{reg_values}\n\n")

        # only update registers whose values have been changed
        self.__update_registers_from_dict(reg_values)



"""' SDK publishing """

import setuptools

with open("README.md", "r", encoding="utf-8") as file:
    long_description = file.read()

setuptools.setup(
    name="edgepi-python-sdk",
    version="1.1.7",
    author="S.Park",
    author_email="spark@osensa.com",
    description="EdgePi Python SDK package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EdgePi-Cloud/edgepi-python-sdk",
    project_urls={
        "Bug Tracker": "https://github.com/EdgePi-Cloud/edgepi-python-sdk/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(where="src", exclude=["test_edgepi"]),
    package_dir={"": "src"},
    python_requires=">=3.6",
    install_requires=["python-periphery >= 2.3.0", "bitstring >= 3.1.9", "protobuf==3.20.*"],
)



""" unit tests for peripherals/i2c.py """

# pylint: disable=C0413
from unittest import mock
from unittest.mock import patch
import sys
if sys.platform != 'linux':
    sys.modules['periphery'] = mock.MagicMock()
import pytest
from edgepi.peripherals.i2c import I2CDevice

# pylint: disable=E1101
if sys.platform != 'linux':
    @pytest.mark.parametrize("fd, mock_expects",
                            [( '/dev/i2c-10', ['/dev/i2c-10']),
                             ( '/dev/i2c-11', ['/dev/i2c-11']),
                             ( '/dev/i2c-12', ['/dev/i2c-12']),
                             ( '/dev/i2c-13', ['/dev/i2c-13'])
                            ])
    @patch('edgepi.peripherals.i2c.I2C')
    def test_i2c_init_param(i2c_mock, fd, mock_expects):
        i2c_mock.fd = '/dev/i2c-10'
        i2c_mock.fd = mock_expects[0]
        i2c_device = I2CDevice(fd)
        assert i2c_device.fd == fd

    @pytest.mark.parametrize("addrs, msg, result",
                            [( 30, [31, 32], [[30], False, [31, 32], True]),
                             ( 20, [21, 22], [[20], False, [21, 22], True]),
                             ( 10, [11, 12], [[10], False, [11, 12], True]),
                             ( 0, [1, 2], [[0], False, [1, 2], True]),
                             ( [1,2], [3, 4], [[1,2], False, [3, 4], True])
                            ])
    @patch('edgepi.peripherals.i2c.I2C')
    def test_i2c_set_read_msg(i2c_mock, addrs, msg, result):
        i2c_mock.fd = '/dev/i2c-10'
        i2c_dev = I2CDevice('/dev/i2c-10')
        i2c_dev.i2cdev.Message.return_value.data = i2c_dev.i2cdev.Message
        list_msg = i2c_dev.set_read_msg(addrs, msg)
        i2c_dev.i2cdev.Message.assert_any_call(result[0], read=result[1])
        i2c_dev.i2cdev.Message.assert_any_call(result[2], read=result[3])
        assert len(list_msg) == i2c_dev.i2cdev.Message.call_count


    # Get Read Message
    # Set Write Message
    @pytest.mark.parametrize("addrs, msg, result",
                            [( 30, [31, 32], [[30, 31, 32], False]),
                             ( 20, [21, 22], [[20, 21, 22], False]),
                             ( 10, [11, 12], [[10, 11, 12], False]),
                             ( 0, [1, 2], [[0, 1, 2], False]),
                             ( [3,4], [1, 2], [[3, 4, 1, 2], False])
                            ])
    @patch('edgepi.peripherals.i2c.I2C')
    def test_i2c_set_write_msg(i2c_mock, addrs, msg, result):
        i2c_mock.fd = '/dev/i2c-10'
        i2c_device = I2CDevice('/dev/i2c-10')
        i2c_device.i2cdev.Message.return_value.data = i2c_device.i2cdev.Message
        list_msg = i2c_device.set_write_msg(addrs, msg)
        i2c_device.i2cdev.Message.assert_called_with(result[0], read=result[1])
        assert len(list_msg) == i2c_device.i2cdev.Message.call_count

else:
    @pytest.mark.parametrize("fd, result",
                            [( '/dev/i2c-10', ['/dev/i2c-10'])
                            ])
    @patch('edgepi.peripherals.i2c.I2C')
    def test_i2c_init_param(i2c_mock,fd, result):
        i2c_mock.fd = '/dev/i2c-10'
        i2c_device = I2CDevice(fd)
        assert i2c_device.fd == result[0]


    @pytest.mark.parametrize("addrs, msg, result",
                            [( 30, [31, 32], [[30], False, [31, 32], True]),
                             ( 20, [21, 22], [[20], False, [21, 22], True]),
                             ( 10, [11, 12], [[10], False, [11, 12], True]),
                             ( 0, [1, 2], [[0], False, [1, 2], True])
                            ])
    @patch('edgepi.peripherals.i2c.I2C')
    def test_i2c_set_read_msg(i2c_mock, addrs, msg, result):
        i2c_mock.fd = '/dev/i2c-10'
        i2c_device = I2CDevice('/dev/i2c-10')
        i2c_device.i2cdev.Message.return_value.data = i2c_device.i2cdev.Message
        list_msg = i2c_device.set_read_msg(addrs, msg)
        i2c_device.i2cdev.Message.assert_any_call(result[0], read=result[1])
        i2c_device.i2cdev.Message.assert_any_call(result[2], read=result[3])
        assert len(list_msg) == i2c_device.i2cdev.Message.call_count


    # Get Read Message
    # Set Write Message
    @pytest.mark.parametrize("addrs, msg, result",
                            [( 30, [31, 32], [[30, 31, 32], False]),
                             ( 20, [21, 22], [[20, 21, 22], False]),
                             ( 10, [11, 12], [[10, 11, 12], False]),
                             ( 0, [1, 2], [[0, 1, 2], False])
                            ])
    @patch('edgepi.peripherals.i2c.I2C')
    def test_i2c_set_write_msg(i2c_mock, addrs, msg, result):
        i2c_mock.fd = '/dev/i2c-10'
        i2c_device = I2CDevice('/dev/i2c-10')
        i2c_device.i2cdev.Message.return_value.data = i2c_device.i2cdev.Message
        list_msg = i2c_device.set_write_msg(addrs, msg)
        i2c_device.i2cdev.Message.assert_called_with(result[0], read=result[1])
        assert len(list_msg) == i2c_device.i2cdev.Message.call_count

    #Transfer Read
    @pytest.mark.parametrize("dev_address, msg, result",
                            [( 33, [2,False,[0],True], [255]),
                             ( 32, [3,False,[0],True], [255]),
                             ( 32, [6,False,[0],True], [255]),
                             ( 33, [7,False,[0],True], [255]),
                            ])
    @patch('edgepi.peripherals.i2c.I2C')
    @patch('edgepi.peripherals.i2c.I2CDevice.transfer')
    def test_i2c_transfer_read(i2c_transfer_mock, i2c_mock, dev_address, msg, result):
        i2c_mock.fd = '/dev/i2c-10'
        i2c_device = I2CDevice('/dev/i2c-10')
        i2c_transfer_mock.return_value = 255
        list_msg = i2c_device.set_read_msg(msg[0], msg[2])
        list_msg = i2c_device.transfer(dev_address, list_msg)
        assert list_msg == result[0]



"""tests for edgepi_tc module"""


from copy import deepcopy
from unittest import mock
from unittest.mock import call, patch
import sys

sys.modules['periphery'] = mock.MagicMock()

# pylint: disable=wrong-import-position

import pytest
from edgepi.tc.edgepi_tc import EdgePiTC
from edgepi.tc.tc_constants import (
    AvgMode,
    CJHighMask,
    CJLowMask,
    CJMode,
    ConvMode,
    DecBits4,
    DecBits6,
    FaultMode,
    NoiseFilterMode,
    OpenCircuitMode,
    OpenMask,
    OvuvMask,
    TCAddresses,
    TCHighMask,
    TCLowMask,
    TCType,
    VoltageMode,
)
from edgepi.tc.tc_faults import FaultMsg, FaultType, Fault

# pylint: disable=protected-access



@pytest.fixture(name="tc")
def fixture_test_edgepi_tc(mocker):
    # mocker acts as context manager
    mocker.patch("edgepi.peripherals.spi.SPI")
    mocker.patch("edgepi.tc.edgepi_tc.Bits")
    # yield instead of return so local state (i.e mocks) not lost
    # upon returning EdgePiTC object in test functions below
    yield EdgePiTC()


@pytest.mark.parametrize(
    "reg_address, data",
    [
        (TCAddresses.CR0_R.value, [TCAddresses.CR0_R.value, 0xFF]),
        (TCAddresses.CR0_W.value, [TCAddresses.CR0_W.value, 0xFF]),
        (TCAddresses.CR1_R.value, [TCAddresses.CR1_R.value, 0xFF]),
        (TCAddresses.CR1_W.value, [TCAddresses.CR1_W.value, 0xFF]),
    ],
)
@patch("edgepi.peripherals.spi.SpiDevice.transfer")
def test_read_register_passes_data(mock_transfer, reg_address, data, tc):
    tc._EdgePiTC__read_register(reg_address)
    mock_transfer.assert_called_once_with(data)


@pytest.mark.parametrize(
    "reg_address, out",
    [
        (TCAddresses.CR0_R.value, [TCAddresses.CR0_R.value, 0x0]),
        (TCAddresses.CR0_W.value, [TCAddresses.CR0_W.value, 0x0]),
        (TCAddresses.CR1_R.value, [TCAddresses.CR1_R.value, 0x3]),
        (TCAddresses.CR1_W.value, [TCAddresses.CR1_W.value, 0x3]),
    ],
)
def test_read_register_returns_data(mocker, reg_address, out, tc):
    mocker.patch("edgepi.peripherals.spi.SpiDevice.transfer", return_value=out)
    out_data = tc._EdgePiTC__read_register(reg_address)
    assert out_data == out


@pytest.mark.parametrize(
    "reg_address, value",
    [
        (TCAddresses.CR0_W.value, 0xFF),
        (TCAddresses.CR1_W.value, 0xFF),
    ],
)
@patch("edgepi.peripherals.spi.SpiDevice.transfer")
def test_write_to_register_passes_data(mock_transfer, reg_address, value, tc):
    tc._EdgePiTC__read_register(reg_address)
    data = [reg_address] + [value]
    mock_transfer.assert_called_once_with(data)


@pytest.mark.parametrize(
    "filter_at_fault, pre_filter_map, expected",
    [
        (
            True,
            {FaultType.OPEN: Fault(FaultType.OPEN, FaultMsg.OPEN_OK_MSG, False, True)},
            {},
        ),  # no faults, return only asserting faults
        (
            False,
            {FaultType.OPEN: Fault(FaultType.OPEN, FaultMsg.OPEN_OK_MSG, False, True)},
            {FaultType.OPEN: Fault(FaultType.OPEN, FaultMsg.OPEN_OK_MSG, False, True)},
        ),  # no faults, return all faults
        (
            True,
            {
                FaultType.OVUV: Fault(FaultType.OVUV, FaultMsg.OVUV_OK_MSG, True, True),
                FaultType.OPEN: Fault(FaultType.OPEN, FaultMsg.OPEN_OK_MSG, False, True),
            },
            {FaultType.OVUV: Fault(FaultType.OVUV, FaultMsg.OVUV_OK_MSG, True, True)},
        ),  # all faults, return only asserting faults
        (
            False,
            {
                FaultType.OVUV: Fault(FaultType.OVUV, FaultMsg.OVUV_OK_MSG, True, True),
                FaultType.OPEN: Fault(FaultType.OPEN, FaultMsg.OPEN_OK_MSG, False, True),
            },
            {
                FaultType.OVUV: Fault(FaultType.OVUV, FaultMsg.OVUV_OK_MSG, True, True),
                FaultType.OPEN: Fault(FaultType.OPEN, FaultMsg.OPEN_OK_MSG, False, True),
            },
        ),  # all faults, return all faults
    ],
)
def test_read_faults_filters(mocker, filter_at_fault, pre_filter_map, expected, tc):
    mocker.patch("edgepi.tc.edgepi_tc.Bits")
    with patch("edgepi.tc.edgepi_tc.map_fault_status", return_value=pre_filter_map):
        result = tc.read_faults(filter_at_fault=filter_at_fault)
        assert result == expected


default_reg_values = {
    TCAddresses.CR0_W.value: 0x0,
    TCAddresses.CR1_W.value: 0x0,
    TCAddresses.MASK_W.value: 0x0,
    TCAddresses.CJHF_W.value: 0x0,
    TCAddresses.CJLF_W.value: 0x0,
    TCAddresses.LTHFTH_W.value: 0x0,
    TCAddresses.LTHFTL_W.value: 0x0,
    TCAddresses.LTLFTH_W.value: 0x0,
    TCAddresses.LTLFTL_W.value: 0x0,
    TCAddresses.CJTO_W.value: 0x0,
    TCAddresses.CJTH_W.value: 0x0,
    TCAddresses.CJTL_W.value: 0x0,
}


@pytest.mark.parametrize(
    "args, set_reg_value, updated_regs",
    [
        (
            {"conversion_mode": ConvMode.SINGLE},
            {TCAddresses.CR0_W.value: 0x80},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"conversion_mode": ConvMode.AUTO},
            None,
            {TCAddresses.CR0_W.value: 0x80},
        ),
        (
            {"open_circuit_mode": OpenCircuitMode.HIGH_INPUT_IMPEDANCE},
            None,
            {TCAddresses.CR0_W.value: 0x30},
        ),
        (
            {"open_circuit_mode": OpenCircuitMode.MED_INPUT_IMPEDANCE},
            None,
            {TCAddresses.CR0_W.value: 0x20},
        ),
        (
            {"open_circuit_mode": OpenCircuitMode.LOW_INPUT_IMPEDANCE},
            None,
            {TCAddresses.CR0_W.value: 0x10},
        ),
        (
            {"open_circuit_mode": OpenCircuitMode.DISABLED},
            {TCAddresses.CR0_W.value: 0x10},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"cold_junction_mode": CJMode.DISABLE},
            None,
            {TCAddresses.CR0_W.value: 0x08},
        ),
        (
            {"cold_junction_mode": CJMode.ENABLE},
            {TCAddresses.CR0_W.value: 0x08},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"fault_mode": FaultMode.INTERRUPT},
            None,
            {TCAddresses.CR0_W.value: 0x04},
        ),
        (
            {"fault_mode": FaultMode.COMPARATOR},
            {TCAddresses.CR0_W.value: 0x04},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"noise_filter_mode": NoiseFilterMode.HZ_60},
            {TCAddresses.CR0_W.value: 0x01},
            {TCAddresses.CR0_W.value: 0x00},
        ),
        (
            {"noise_filter_mode": NoiseFilterMode.HZ_50},
            None,
            {TCAddresses.CR0_W.value: 0x01},
        ),
        (
            {"average_mode": AvgMode.AVG_16},
            None,
            {TCAddresses.CR1_W.value: 0x40},
        ),
        (
            {"average_mode": AvgMode.AVG_8},
            None,
            {TCAddresses.CR1_W.value: 0x30},
        ),
        (
            {"average_mode": AvgMode.AVG_4},
            None,
            {TCAddresses.CR1_W.value: 0x20},
        ),
        (
            {"average_mode": AvgMode.AVG_2},
            None,
            {TCAddresses.CR1_W.value: 0x10},
        ),
        (
            {"average_mode": AvgMode.AVG_1},
            {TCAddresses.CR1_W.value: 0x10},
            {TCAddresses.CR1_W.value: 0x00},
        ),
        (
            {"tc_type": TCType.TYPE_B},
            None,
            {TCAddresses.CR1_W.value: 0x00},
        ),
        (
            {"tc_type": TCType.TYPE_E},
            None,
            {TCAddresses.CR1_W.value: 0x01},
        ),
        (
            {"tc_type": TCType.TYPE_J},
            None,
            {TCAddresses.CR1_W.value: 0x02},
        ),
        (
            {"tc_type": TCType.TYPE_K},
            None,
            {TCAddresses.CR1_W.value: 0x03},
        ),
        (
            {"tc_type": TCType.TYPE_N},
            None,
            {TCAddresses.CR1_W.value: 0x04},
        ),
        (
            {"tc_type": TCType.TYPE_R},
            None,
            {TCAddresses.CR1_W.value: 0x05},
        ),
        (
            {"tc_type": TCType.TYPE_S},
            None,
            {TCAddresses.CR1_W.value: 0x06},
        ),
        (
            {"tc_type": TCType.TYPE_T},
            None,
            {TCAddresses.CR1_W.value: 0x07},
        ),
        (
            {"voltage_mode": VoltageMode.GAIN_8},
            None,
            {TCAddresses.CR1_W.value: 0x08},
        ),
        (
            {"voltage_mode": VoltageMode.GAIN_32},
            None,
            {TCAddresses.CR1_W.value: 0x0C},
        ),
        (
            {"cj_high_mask": CJHighMask.CJHIGH_MASK_OFF},
            {TCAddresses.MASK_W.value: 0x20},
            {TCAddresses.MASK_W.value: 0x00},
        ),
        (
            {"cj_high_mask": CJHighMask.CJHIGH_MASK_ON},
            None,
            {TCAddresses.MASK_W.value: 0x20},
        ),
        (
            {"cj_low_mask": CJLowMask.CJLOW_MASK_OFF},
            {TCAddresses.MASK_W.value: 0x10},
            {TCAddresses.MASK_W.value: 0x00},
        ),
        (
            {"cj_low_mask": CJLowMask.CJLOW_MASK_ON},
            None,
            {TCAddresses.MASK_W.value: 0x10},
        ),
        (
            {"tc_high_mask": TCHighMask.TCHIGH_MASK_OFF},
            {TCAddresses.MASK_W.value: 0x08},
            {TCAddresses.MASK_W.value: 0x00},
        ),
        (
            {"tc_high_mask": TCHighMask.TCHIGH_MASK_ON},
            None,
            {TCAddresses.MASK_W.value: 0x08},
        ),
        (
            {"tc_low_mask": TCLowMask.TCLOW_MASK_OFF},
            {TCAddresses.MASK_W.value: 0x04},
            {TCAddresses.MASK_W.value: 0x00},
        ),
        (
            {"tc_low_mask": TCLowMask.TCLOW_MASK_ON},
            None,
            {TCAddresses.MASK_W.value: 0x04},
        ),
        (
            {"ovuv_mask": OvuvMask.OVUV_MASK_OFF},
            {TCAddresses.MASK_W.value: 0x02},
            {TCAddresses.MASK_W.value: 0x00},
        ),
        (
            {"ovuv_mask": OvuvMask.OVUV_MASK_ON},
            None,
            {TCAddresses.MASK_W.value: 0x02},
        ),
        (
            {"open_mask": OpenMask.OPEN_MASK_OFF},
            {TCAddresses.MASK_W.value: 0x01},
            {TCAddresses.MASK_W.value: 0x00},
        ),
        (
            {"open_mask": OpenMask.OPEN_MASK_ON},
            None,
            {TCAddresses.MASK_W.value: 0x01},
        ),
        (
            {"cj_high_threshold": 100},
            {TCAddresses.CR1_W.value: 0x03},
            {TCAddresses.CJHF_W.value: 0x64},
        ),
        ({"cj_low_threshold": -16}, None, {TCAddresses.CJLF_W.value: 0x90}),
        (
            {"lt_high_threshold": 1000, "lt_high_threshold_decimals": DecBits4.P0_9375},
            {TCAddresses.CR1_W.value: 0x03},
            {TCAddresses.LTHFTH_W.value: 0x3E, TCAddresses.LTHFTL_W.value: 0x8F},
        ),
        (
            {"lt_low_threshold": -55, "lt_low_threshold_decimals": DecBits4.P0_9375},
            {TCAddresses.CR1_W.value: 0x03},
            {TCAddresses.LTLFTH_W.value: 0x83, TCAddresses.LTLFTL_W.value: 0x7F},
        ),
        (
            {"cj_offset": 4, "cj_offset_decimals": DecBits4.P0_9375},
            {TCAddresses.CR1_W.value: 0x03},
            {
                TCAddresses.CJTO_W.value: 0x4F,
            },
        ),
        (
            {"cj_temp": 100, "cj_temp_decimals": DecBits6.P0_984375},
            {TCAddresses.CR1_W.value: 0x03},
            {TCAddresses.CJTH_W.value: 0x64, TCAddresses.CJTL_W.value: 0xFC},
        ),
    ],
)
def test_set_config(mocker, args, set_reg_value, updated_regs, tc):
    # need to deepcopy in order to keep default_reg_values unaltered between test cases
    start_values = deepcopy(default_reg_values)

    # if needed, modify default register values for this test case
    if set_reg_value is not None:
        for addx, value in set_reg_value.items():
            start_values[addx] = value

    # need 2nd deepcopy for comparing updated values to start values below
    # otherwise, the modification of values above is counted as an update
    # i.e. keep start_values unaltered by code below
    reg_values = deepcopy(start_values)

    # insert mock register values -- set_config will update these with opcodes below
    mocker.patch(
        "edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__read_registers_to_map", return_value=reg_values
    )
    # mocked out SPI, so it can't read tc_type: need to return it manually
    mocker.patch("edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__get_tc_type", return_value=TCType.TYPE_K)

    tc.set_config(**args)

    for addx, entry in reg_values.items():
        # check registers not updated have not been changed
        if addx not in updated_regs:
            assert not entry["is_changed"]
            assert entry["value"] == start_values[addx]
        # check updates were applied
        else:
            assert reg_values[addx]["value"] == updated_regs[addx]
            assert reg_values[addx]["is_changed"]


@pytest.mark.parametrize(
    "reg_values",
    [
        ({TCAddresses.CR0_W.value: {"is_changed": True, "value": 0xFF}}),
        (
            {
                TCAddresses.CR0_W.value: {"is_changed": True, "value": 0xFF},
                TCAddresses.CR1_W.value: {"is_changed": True, "value": 0xFF},
            }
        ),
        ({TCAddresses.CR0_W.value: {"is_changed": False, "value": 0xFF}}),
        (
            {
                TCAddresses.CR0_W.value: {"is_changed": True, "value": 0xFF},
                TCAddresses.CR1_W.value: {"is_changed": True, "value": 0xFF},
            }
        ),
        (
            {
                TCAddresses.CR0_W.value: {"is_changed": True, "value": 0xFF},
                TCAddresses.CR1_W.value: {"is_changed": True, "value": 0xFF},
                TCAddresses.CJHF_W.value: {"is_changed": False, "value": 0x0},
            }
        ),
    ],
)
@patch("edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__write_to_register")
def test_update_registers_from_dict(mock_write, reg_values, tc):
    # construct list of expected calls to __write_to_register
    write_calls = []
    for addx, entry in reg_values.items():
        if entry["is_changed"]:
            write_calls.append(call(addx, entry["value"]))

    tc._EdgePiTC__update_registers_from_dict(reg_values)

    # assert __write_to_register called with expected calls
    mock_write.assert_has_calls(write_calls, any_order=True)


@pytest.mark.parametrize(
    "cj_temp, cj_temp_decimals",
    [
        (-50, DecBits4.P0_125),
        (-50, DecBits4.P0_125),
    ],
)
def test_overwrite_cold_junction_temp(cj_temp, cj_temp_decimals, tc):
    with patch("edgepi.tc.edgepi_tc.EdgePiTC.set_config") as mock_set_config:
        tc.overwrite_cold_junction_temp(cj_temp, cj_temp_decimals)
        args = {"cj_temp": cj_temp, "cj_temp_decimals": cj_temp_decimals}
        mock_set_config.assert_called_once_with(**args)


@pytest.mark.parametrize(
    "cr0_val, cmd",
    [
        (0x00, 0x40),
        (0b10111101, 0b11111101),
    ],
)
def test_single_sample(mocker, cr0_val, cmd, tc):
    mocker.patch("edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__read_register", return_value=[0, cr0_val])
    mocker.patch("edgepi.tc.edgepi_tc.calc_conv_time", return_value=200)
    with patch("edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__write_to_register") as mock_write:
        tc.single_sample()
        mock_write.assert_called_once_with(TCAddresses.CR0_W.value, cmd)


@pytest.mark.parametrize(
    "cr0_val, cmd",
    [
        (0x00, 0x02),
        (0b10111101, 0b10111111),
    ],
)
def test_clear_faults(mocker, cr0_val, cmd, tc):
    mocker.patch("edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__read_register", return_value=[0, cr0_val])
    with patch("edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__write_to_register") as mock_write:
        tc.clear_faults()
        mock_write.assert_called_once_with(TCAddresses.CR0_W.value, cmd)


@pytest.mark.parametrize(
    "cr1_val, tc_type",
    [
        (0x00, TCType.TYPE_B),
        (0x01, TCType.TYPE_E),
        (0x02, TCType.TYPE_J),
        (0x03, TCType.TYPE_K),
        (0x04, TCType.TYPE_N),
        (0x05, TCType.TYPE_R),
        (0x06, TCType.TYPE_S),
        (0x07, TCType.TYPE_T),
        (0xFF, None),
    ],
)
def test_get_tc_type(mocker, cr1_val, tc_type):
    # not using fixture because need Bits not mocked here
    mocker.patch("edgepi.peripherals.spi.SPI")
    tc = EdgePiTC()
    mocker.patch("edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__read_register", return_value=[0, cr1_val])
    assert tc._EdgePiTC__get_tc_type() == tc_type


@patch("edgepi.tc.edgepi_tc.EdgePiTC._EdgePiTC__write_to_register")
def test_tc_reset_registers(mock_write, tc):
    write_calls = [call(addx, value) for addx, value in tc.default_reg_values.items()]
    tc.reset_registers()
    mock_write.assert_has_calls(write_calls, any_order=True)



"""unit tests for edgepi_dout module"""

# pylint: disable=C0413

from unittest import mock
import sys
if sys.platform != 'linux':
    sys.modules['periphery'] = mock.MagicMock()

from contextlib import nullcontext as does_not_raise
import pytest
from edgepi.gpio.gpio_constants import GpioPins
from edgepi.gpio.gpio_configs import DOUTPins
from edgepi.digital_output.edgepi_digital_output import EdgePiDigitalOutput, InvalidPinName

@pytest.mark.parametrize("pin_name, state, error",
                        [(GpioPins.DOUT1, True, does_not_raise()),
                         (GpioPins.DOUT2, True, does_not_raise()),
                         (GpioPins.DOUT3, True, does_not_raise()),
                         (GpioPins.DOUT4, True, does_not_raise()),
                         (GpioPins.DOUT5, True, does_not_raise()),
                         (GpioPins.DOUT6, True, does_not_raise()),
                         (GpioPins.DOUT7, True, does_not_raise()),
                         (GpioPins.DOUT8, True, does_not_raise()),
                         (GpioPins.DOUT1, False, does_not_raise()),
                         (GpioPins.DOUT2, False, does_not_raise()),
                         (GpioPins.DOUT3, False, does_not_raise()),
                         (GpioPins.DOUT4, False, does_not_raise()),
                         (GpioPins.DOUT5, False, does_not_raise()),
                         (GpioPins.DOUT6, False, does_not_raise()),
                         (GpioPins.DOUT7, False, does_not_raise()),
                         (GpioPins.DOUT8, False, does_not_raise()),
                         (GpioPins.DOUT8, None, pytest.raises(ValueError)),
                         (None, False, pytest.raises(InvalidPinName)),
                         (GpioPins.DIN1, False, pytest.raises(InvalidPinName))])
def test_edgepi_digital_output_state(mocker, pin_name, state, error):
    gpio_chip=mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIO.write_gpio_pin_state")
    expander_set = mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIO.set_expander_pin")
    expander_clear = mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIO.clear_expander_pin")
    dout = EdgePiDigitalOutput()
    with error:
        dout.digital_output_state(pin_name, state)
        if pin_name.value in [DOUTPins.DOUT1.value,DOUTPins.DOUT2.value]:
            gpio_chip.assert_called_once_with(pin_name.value, state)
        else:
            if state:
                expander_set.assert_called_once_with(pin_name.value)
            else:
                expander_clear.assert_called_once_with(pin_name.value)

@pytest.mark.parametrize("pin_name, direction, error",
                        [(GpioPins.DOUT1, True, does_not_raise()),
                         (GpioPins.DOUT2, True, does_not_raise()),
                         (GpioPins.DOUT3, True, does_not_raise()),
                         (GpioPins.DOUT4, True, does_not_raise()),
                         (GpioPins.DOUT5, True, does_not_raise()),
                         (GpioPins.DOUT6, True, does_not_raise()),
                         (GpioPins.DOUT7, True, does_not_raise()),
                         (GpioPins.DOUT8, True, does_not_raise()),
                         (GpioPins.DOUT1, False, does_not_raise()),
                         (GpioPins.DOUT2, False, does_not_raise()),
                         (GpioPins.DOUT3, False, does_not_raise()),
                         (GpioPins.DOUT4, False, does_not_raise()),
                         (GpioPins.DOUT5, False, does_not_raise()),
                         (GpioPins.DOUT6, False, does_not_raise()),
                         (GpioPins.DOUT7, False, does_not_raise()),
                         (GpioPins.DOUT8, False, does_not_raise()),
                         (GpioPins.DOUT8, None, pytest.raises(ValueError)),
                         (None, False, pytest.raises(InvalidPinName)),
                         (GpioPins.DIN1, False, pytest.raises(InvalidPinName))])
def test_edgepi_digital_output_direction(mocker, pin_name, direction, error):
    gpio_chip=mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIO.set_gpio_pin_dir")
    exp_dir_in = mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIO.set_expander_pin_direction_in")
    exp_dir_out = mocker.patch("edgepi.gpio.edgepi_gpio.EdgePiGPIO.set_expander_pin_direction_out")
    dout = EdgePiDigitalOutput()
    with error:
        dout.digital_output_direction(pin_name, direction)
        if pin_name.value in [DOUTPins.DOUT1.value,DOUTPins.DOUT2.value]:
            gpio_chip.assert_called_once_with(pin_name.value, direction)
        else:
            if direction:
                exp_dir_in.assert_called_once_with(pin_name.value)
            else:
                exp_dir_out.assert_called_once_with(pin_name.value)



""" User interface for EdgePi ADC """


from dataclasses import dataclass
from enum import Enum
from operator import attrgetter
import logging
import time


from bitstring import pack
from edgepi.adc.adc_query_lang import PropertyValue, ADCProperties
from edgepi.calibration.calibration_constants import CalibParam
from edgepi.peripherals.spi import SpiDevice as SPI
from edgepi.adc.adc_commands import ADCCommands
from edgepi.adc.adc_constants import (
    ADC1DataRate,
    ADC2DataRate,
    ADCChannel as CH,
    ADCComs,
    ADCNum,
    ADCPower,
    ConvMode,
    ADCReg,
    DifferentialPair,
    FilterMode,
    ADCReferenceSwitching,
    ADC_NUM_REGS,
    ADC_VOLTAGE_READ_LEN,
    CheckMode,
    DiffMode,
    IDACMUX,
    IDACMAG,
    REFMUX,
    RTDModes,
    AllowedChannels,
)
from edgepi.adc.adc_voltage import code_to_voltage, check_crc, code_to_temperature
from edgepi.gpio.edgepi_gpio import EdgePiGPIO
from edgepi.gpio.gpio_configs import ADCPins
from edgepi.utilities.utilities import filter_dict
from edgepi.reg_helper.reg_helper import OpCode, apply_opcodes
from edgepi.adc.adc_multiplexers import (
    generate_mux_opcodes,
    validate_channels_allowed,
)
from edgepi.adc.adc_conv_time import expected_initial_time_delay, expected_continuous_time_delay
from edgepi.adc.adc_status import get_adc_status
from edgepi.calibration.edgepi_eeprom import EdgePiEEPROM

_logger = logging.getLogger(__name__)


@dataclass
class ADCReadFields:
    """
    ADC state properties specific to each of ADC1 and ADC2
    """

    conversion_mode: PropertyValue
    data_rate: PropertyValue
    mux_p: PropertyValue
    mux_n: PropertyValue


class ADCState:
    """ADC state intended for reading by users"""

    def __init__(self, reg_map: dict):
        self.__reg_map = reg_map
        self.adc_1: ADCReadFields = ADCReadFields(
            self.__get_state(ADCProperties.CONV_MODE),
            self.__get_state(ADCProperties.DATA_RATE_1),
            self.__get_state(ADCProperties.ADC1_MUXP),
            self.__get_state(ADCProperties.ADC1_MUXN),
        )
        self.adc_2: ADCReadFields = ADCReadFields(
            PropertyValue("continuous", ADCProperties.CONV_MODE),
            self.__get_state(ADCProperties.DATA_RATE_2),
            self.__get_state(ADCProperties.ADC2_MUXP),
            self.__get_state(ADCProperties.ADC2_MUXN),
        )
        self.filter_mode: PropertyValue = self.__get_state(ADCProperties.FILTER_MODE)
        self.status_byte: PropertyValue = self.__get_state(ADCProperties.STATUS_MODE)
        self.checksum_mode: PropertyValue = self.__get_state(ADCProperties.CHECK_MODE)
        self.rtd_on: bool = self.__get_rtd_state()

    def __query_state(self, adc_property: ADCProperties) -> PropertyValue:
        """
        Read the current state of configurable ADC properties

        Args:
            `adc_property` (ADCProperties): ADC property whose state is to be read
            `reg_map`: register map formatted as {addx (int): value (int)}

        Returns:
            `PropertyValue`: information about the current value of this property
        """
        # value of this adc_property's register
        reg_value = self.__reg_map[adc_property.value.addx]

        # get value of bits corresponding to this property by letting through only the bits
        # that were "masked" when setting this property (clear all bits except the property bits)
        adc_property_bits = (~adc_property.value.mask) & reg_value

        # name of current value of this adc_property
        adc_property_value = adc_property.value.values[adc_property_bits]
        _logger.debug(
            (
                f"query_state: query_property='{adc_property}',"
                " adc_property_bits={hex(adc_property_bits)},"
                f" adc_property_value='{adc_property_value}'"
            )
        )

        return adc_property_value

    def __get_state(self, adc_property: ADCProperties) -> PropertyValue:
        """
        Read the current state of configurable ADC properties

        Args:
            `adc_property` (ADCProperties): ADC property whose state is to be read

        Returns:
            `PropertyValue`: information about the current value of this mode
        """
        return self.__query_state(adc_property)

    def __get_current_rtd_state(self) -> dict[str, PropertyValue]:
        return {
            "adc_1_analog_in": self.adc_1.mux_p.code,
            "adc_1_mux_n": self.adc_1.mux_n.code,
            "idac_1_mux": self.__get_state(ADCProperties.IDAC1_MUX).code,
            "idac_2_mux": self.__get_state(ADCProperties.IDAC2_MUX).code,
            "idac_1_mag": self.__get_state(ADCProperties.IDAC1_MAG).code,
            "idac_2_mag": self.__get_state(ADCProperties.IDAC2_MAG).code,
            "pos_ref_inp": self.__get_state(ADCProperties.REFMUX_POS).code,
            "neg_ref_inp": self.__get_state(ADCProperties.REFMUX_NEG).code,
        }

    def __get_rtd_state(self) -> bool:
        """
        Get on/off RTD state.
        """
        return self.__get_current_rtd_state() == RTDModes.RTD_ON.value


class ADCStateMissingMap(Exception):
    """ "Raised if ADCState.get_map() is called before ADCState has been assigned a cached state"""


class ADCRegisterUpdateError(Exception):
    """Raised when a register update fails to set register to expected value"""


class VoltageReadError(Exception):
    """Raised if a voltage read fails to return the expected number of bytes"""


class ContinuousModeError(Exception):
    """Raised when `read_voltage` is called and ADC is not in CONTINUOUS conversion mode"""


class RTDEnabledError(Exception):
    """Raised when user attempts to set ADC configuration that conflicts with RTD mode"""


class InvalidDifferentialPairError(Exception):
    """
    Raised when calibration values are attempted to be retrieved for an invalid Differential
    read pair
    """


class CalibKeyMissingError(Exception):
    """Raised when calibration values are missing from EEPROM dictionary"""


class EdgePiADC(SPI):
    """
    EdgePi ADC device

    Warning, when caching is enabled to track the ADC's internal state, EdgePiADC objects are
    safe to use only within a single dev environment. Using multiple EdgePiADC objects,
    each within a different environment, will lead to the cached ADC state being out of sync
    with the actual hardware state. To avoid this, disable caching when creating the
    EdgePiADC object; the state of the ADC will be read from hardware instead, at the
    cost of increased SPI reading load.
    """

    # keep track of ADC register map state for state caching
    __state: dict = {}

    # default RTD model-dependent hardware constants
    RTD_SENSOR_RESISTANCE = 100 # RTD sensor resistance value (Ohms)
    RTD_SENSOR_RESISTANCE_VARIATION = 0.385 # RTD sensor resistance variation (Ohms/°C)

    # TODO: this should be part of eeprom_data. Retrieve from eeprom_data in calling
    # functions when available
    r_ref = 1326.20

    def __init__(
        self,
        enable_cache: bool = False,
        rtd_sensor_resistance: float = None,
        rtd_sensor_resistance_variation: float = None
        ):
        """
        Args:
            `enable_cache` (bool): set to True to enable state-caching

            `rtd_sensor_resistance` (float): set RTD material-dependent resistance value (Ohms)

            `rtd_sensor_resistance_variation` (float): set RTD model-dependent resistance
                variation (Ohms/°C)
        """

        super().__init__(bus_num=6, dev_id=1)
        # declare instance vars before config call below
        self.enable_cache = enable_cache

        # Load eeprom data and generate dictionary of calibration dataclass
        eeprom = EdgePiEEPROM()
        eeprom_data  = eeprom.get_edgepi_reserved_data()
        self.adc_calib_params = eeprom_data.adc_calib_parms

        self.adc_ops = ADCCommands()
        self.gpio = EdgePiGPIO()
        # ADC always needs to be in CRC check mode. This also updates the internal __state.
        # If this call to __config is removed, replace with a call to get_register_map to
        # initialize __state.
        self.__config(checksum_mode=CheckMode.CHECK_BYTE_CRC)
        # TODO: adc reference should ba a config that customer passes depending on the range of
        # voltage they are measuring. To be changed later when range config is implemented
        self.set_adc_reference(ADCReferenceSwitching.GND_SW1.value)

        # user updated rtd hardware constants
        self.rtd_sensor_resistance = (
            rtd_sensor_resistance
            if rtd_sensor_resistance is not None
            else EdgePiADC.RTD_SENSOR_RESISTANCE
        )

        self.rtd_sensor_resistance_variation = (
            rtd_sensor_resistance_variation
            if rtd_sensor_resistance_variation is not None
            else EdgePiADC.RTD_SENSOR_RESISTANCE_VARIATION
        )

    def __reapply_config(self):
        """
        Restore ADC to custom EdgePi configuration
        """
        # turn RTD off to allow updates in case RTD is on
        self.rtd_mode(enable=False)
        self.__config(
            adc_1_analog_in=CH.FLOAT,
            adc_2_analog_in=CH.FLOAT,
            adc_1_mux_n=CH.AINCOM,
            adc_2_mux_n=CH.AINCOM,
            checksum_mode=CheckMode.CHECK_BYTE_CRC,
        )

    def __update_cache_map(self, reg_map: dict):
        """
        Format a register map to the same format as `read_registers_to_map` and
        assign to the internal ADC state.

        Args:
            `reg_map` (dict): register map formatted as {int: {"value": int}}
        """
        EdgePiADC.__state = {addx: entry["value"] for (addx, entry) in reg_map.items()}

    def __get_register_map(
        self,
        override_cache: bool = False,
    ) -> dict[int, int]:
        """
        Get a mapping of register addresses to register values, for the specified
        number of registers. All ADC methods which require register reading should
        call this method to get the register values.

        Args:
            `override_cache` (bool): force update cached state via SPI read

        Returns:
            dict: mapping of uint register addresses to uint register values
        """
        if not EdgePiADC.__state or override_cache or not self.enable_cache:
            EdgePiADC.__state = self.__read_registers_to_map()

        # if caching is disabled, don't use cached state for return (dict() deepcopies)
        return (
            EdgePiADC.__state
            if self.enable_cache
            else dict(EdgePiADC.__state)
        )

    def __read_register(self, start_addx: ADCReg, num_regs: int = 1):
        """
        Read data from ADC registers, either individually or as a block.

        Args:
            `start_addx` (ADCReg): address of register to start read at.

            `num_regs` (int): number of registers to read from start_addx, including
                start_addx register.

        Returns:
            `list`: a list of ints representing bytes formatted as [reg_1_data, reg_2_data,...]
                where reg_1_data is the data of the register at start_addx.
        """
        if num_regs < 1:
            raise ValueError("Number of registers to read must be at least 1")

        code = self.adc_ops.read_register_command(start_addx.value, num_regs)
        _logger.debug(f"ADC __read_register -> data in: {code}")
        out = self.transfer(code)
        _logger.debug(f"ADC __read_register -> data out: {out}")

        # first 2 entries are null bytes
        return out[2:]

    def __write_register(self, start_addx: ADCReg, data: list[int]):
        """
        Write data to ADC registers, either individually or as a block.

        Args:
            `start_addx` (ADCReg): address of register to start writing data from.

            `data` (list): list of int values, where each value represents a byte of data
                to write to a register.
        """
        if len(data) < 1:
            raise ValueError("Number of registers to write to must be at least 1")

        code = self.adc_ops.write_register_command(start_addx.value, data)
        _logger.debug(f"ADC __write_register -> data in: {code}")
        out = self.transfer(code)
        _logger.debug(f"ADC __write_register -> data out: {out}")

        return out

    def set_adc_reference(self, reference_config: ADCReferenceSwitching = None):
        """
        Setting ADC referene terminal state. pin 18 and 23 labeled IN GND on the enclosure. It can
        be configured as regular 0V GND or 12V GND.

        Args:
            reference_config: (ADCReferenceSwitching): selecting none, 1, 2, both
        """
        if reference_config == ADCReferenceSwitching.GND_SW1.value:
            self.gpio.set_pin_state(ADCPins.GNDSW_IN1.value)
            self.gpio.clear_pin_state(ADCPins.GNDSW_IN2.value)
        elif reference_config == ADCReferenceSwitching.GND_SW2.value:
            self.gpio.set_pin_state(ADCPins.GNDSW_IN2.value)
            self.gpio.clear_pin_state(ADCPins.GNDSW_IN1.value)
        elif reference_config == ADCReferenceSwitching.GND_SW_BOTH.value:
            self.gpio.set_pin_state(ADCPins.GNDSW_IN1.value)
            self.gpio.set_pin_state(ADCPins.GNDSW_IN2.value)
        elif reference_config == ADCReferenceSwitching.GND_SW_NONE.value:
            self.gpio.clear_pin_state(ADCPins.GNDSW_IN1.value)
            self.gpio.clear_pin_state(ADCPins.GNDSW_IN2.value)

    def stop_conversions(self, adc_num: ADCNum):
        """
        Halt voltage read conversions when ADC is set to perform continuous conversions
        """
        stop_cmd = self.adc_ops.stop_adc(adc_num=adc_num.value)
        self.transfer(stop_cmd)

    def __send_start_command(self, adc_num: ADCNum):
        """Triggers ADC conversion(s)"""
        start_cmd = self.adc_ops.start_adc(adc_num=adc_num.value)
        self.transfer(start_cmd)

    def start_conversions(self, adc_num: ADCNum):
        """
        Start ADC voltage read conversions. If ADC is continuous conversion mode,
        this method must be called before performing reads. If ADC is in
        pulse conversion mode, this method does not need to be called before
        performing reads.
        """

        # get state for configs relevant to conversion delay
        state = self.get_state()
        conv_mode = state.adc_1.conversion_mode.code
        data_rate = (
            state.adc_1.data_rate.code if adc_num == ADCNum.ADC_1 else state.adc_2.data_rate.code
        )
        filter_mode = state.filter_mode.code

        conv_delay = expected_initial_time_delay(
            adc_num, data_rate.value.op_code, filter_mode.value.op_code
        )
        _logger.debug(
            (
                f"\nComputed time delay = {conv_delay} (ms) with the following config opcodes:\n"
                f"adc_num={adc_num}, conv_mode={hex(conv_mode.value.op_code)}, "
                f"data_rate={hex(data_rate.value.op_code)}"
                f"filter_mode={hex(filter_mode.value.op_code)}\n"
            )
        )
        self.__send_start_command(adc_num)

        # apply delay for first conversion
        time.sleep(conv_delay / 1000)

    def clear_reset_bit(self):
        """
        Clear the ADC reset bit. This reset bits acts as a flag that is raised when the ADC
        resets, which automatically happens during power on, or when a reset command is passed.
        Clearing this bit allows subsequent non-standard resets to be detected (by reading the
        STATUS byte of a voltage read).
        """
        self.__config(reset_clear=ADCPower.RESET_CLEAR)

    def __read_data(self, adc: ADCNum, data_size: int):
        """Sends command to ADC to get new voltage conversion data"""
        return self.transfer([adc.value.read_cmd] + [255] * data_size)

    def __voltage_read(self, adc_num: ADCNum):
        """
        Performs ADC voltage read and formats output into status, voltage,
        and check bytes

        Returns:
            int, list[int], int: uint values of representations of voltage read data ordered as
                (status_byte, voltage_data_bytes, check_byte)
        """
        read_data = self.__read_data(adc_num, ADC_VOLTAGE_READ_LEN)

        if len(read_data) - 1 != ADC_VOLTAGE_READ_LEN:
            raise VoltageReadError(
                f"Voltage read failed: incorrect number of bytes ({len(read_data)}) retrieved"
            )

        status_code = read_data[1]

        voltage_code = read_data[2 : (2 + adc_num.value.num_data_bytes)]

        check_code = read_data[6]

        check_crc(voltage_code, check_code)

        return status_code, voltage_code, check_code

    @staticmethod
    def __get_diff_id(mux_p: PropertyValue, mux_n: PropertyValue) -> int:
        """
        Get differential pair id number for retrieving differential pair calibration values
        """
        #  values are the keys from adc_calib_params
        diff_ids = {
                DiffMode.DIFF_1.value: 8,
                DiffMode.DIFF_2.value: 9,
                DiffMode.DIFF_3.value: 10,
                DiffMode.DIFF_4.value: 11,
            }
        diff_pair = DifferentialPair(mux_p.code, mux_n.code)
        diff_id = diff_ids.get(diff_pair)
        if diff_id is None:
            raise InvalidDifferentialPairError(
                    f"Cannot retrieve calibration values for invalid differential pair {diff_pair}"
                    )
        return diff_id

    def __get_calibration_values(self, adc_calibs: dict, adc_num: ADCNum) -> CalibParam:
        """
        Retrieve voltage reading calibration values based on currently configured
        input multiplexer channels

        Args:
            `adc_calibs` (dict): eeprom adc calibration values
            `adc_num` (ADCNum): adc number voltage is being read from

        Returns:
            `CalibParam`: gain and offset values for voltage reading calibration
        """
        state = self.get_state()
        mux_p = attrgetter(f"adc_{adc_num.value.id_num}.mux_p")(state)
        mux_n = attrgetter(f"adc_{adc_num.value.id_num}.mux_n")(state)

        # assert neither mux is set to float mode
        if CH.FLOAT in (mux_p.code, mux_n.code):
            raise ValueError("Cannot retrieve calibration values for channel in float mode")

        calib_key = mux_p.value if mux_n.code == CH.AINCOM else self.__get_diff_id(mux_p, mux_n)

        calibs = adc_calibs.get(calib_key)

        if calibs is None:
            _logger.error("Failed to find ADC calibration values")
            raise CalibKeyMissingError(
                (
                    "Failed to retrieve calibration values from eeprom dictionary: "
                    f"dict is missing key = {calib_key}"
                    f"\neeprom_calibs = \n{adc_calibs}"
                )
            )

        return calibs

    def __continuous_time_delay(self, adc_num: ADCNum):
        """Compute and enforce continuous conversion time delay"""
        state = self.get_state()
        # get continuous mode time delay and wait here (delay is needed between each conversion)
        data_rate = (
            state.adc_1.data_rate.code if adc_num == ADCNum.ADC_1 else state.adc_2.data_rate.code
        )
        delay = expected_continuous_time_delay(adc_num, data_rate.value.op_code)

        time.sleep(delay / 1000)

    def __check_adc_1_conv_mode(self):
        # assert adc is in continuous mode
        state = self.get_state()
        if state.adc_1.conversion_mode.code != ConvMode.CONTINUOUS:
            raise ContinuousModeError(
                "ADC1 must be in CONTINUOUS conversion mode in order to call this method."
            )

    def read_voltage(self, adc_num: ADCNum):
        """
        Read voltage input to either ADC1 or ADC2, when performing single channel reading
        or differential reading. For ADC1 reading, only use this method when ADC1 is configured
        to `CONTINUOUS` conversion mode. For `PULSE` conversion mode, use `single_sample` instead.

        Args:
            `adc_num` (ADCNum): the ADC to be read

        Returns:
            `float`: input voltage (V) read from the indicated ADC
        """
        if adc_num == ADCNum.ADC_1:
            self.__check_adc_1_conv_mode()

        self.__continuous_time_delay(adc_num)

        status_code, voltage_code, _ = self.__voltage_read(adc_num)

        # log STATUS byte
        status = get_adc_status(status_code)
        _logger.debug(f"Logging STATUS byte:\n{status}")

        calibs = self.__get_calibration_values(self.adc_calib_params, adc_num)

        # convert from code to voltage
        return code_to_voltage(voltage_code, adc_num.value, calibs)


    def read_rtd_temperature(self):
        """
        Read RTD temperature continuously. Note, to obtain valid temperature values,
        RTD mode must first be enabled by calling `rtd_mode`. ADC1 must also be configured
        to `CONTINUOUS` conversion mode via `set_config`, before calling this method.

        Returns:
            `float`: RTD measured temperature (°C)
        """
        self.__check_adc_1_conv_mode()

        self.__continuous_time_delay(ADCNum.ADC_1)

        _, voltage_code, _ = self.__voltage_read(ADCNum.ADC_1)

        # TODO: get RTD calibs from eeprom once added
        return code_to_temperature(
            voltage_code,
            self.r_ref,
            self.rtd_sensor_resistance,
            self.rtd_sensor_resistance_variation
        )

    def __enforce_pulse_mode(self):
        # assert adc is in PULSE mode (check ADCState) -> set to PULSE if not
        state = self.get_state()
        if state.adc_1.conversion_mode.code != ConvMode.PULSE:
            self.__config(conversion_mode=ConvMode.PULSE)

    def single_sample(self):
        """
        Trigger a single ADC1 voltage sampling event, when performing single channel reading or
        differential reading. ADC1 must be in `PULSE` conversion mode before calling this method.
        Do not call this method for voltage reading if ADC is configured to `CONTINUOUS`
        conversion mode: use `read_voltage` instead.

        Returns:
            `float`: input voltage (V) read from ADC1
        """
        self.__enforce_pulse_mode()

        # send command to trigger conversion
        self.start_conversions(ADCNum.ADC_1)

        # send command to read conversion data.
        status_code, voltage_code, _ = self.__voltage_read(ADCNum.ADC_1)

        # log STATUS byte
        status = get_adc_status(status_code)
        _logger.debug(f"Logging STATUS byte:\n{status}")

        calibs = self.__get_calibration_values(self.adc_calib_params, ADCNum.ADC_1)

        # convert from code to voltage
        return code_to_voltage(voltage_code, ADCNum.ADC_1.value, calibs)

    def single_sample_rtd(self):
        """
        Trigger a single RTD temperature sampling event. Note, to obtain valid temperature values,
        RTD mode must first be enabled by calling `rtd_mode`. ADC1 must also be configured
        to `PULSE` conversion mode via `set_config`, before calling this method.

        Returns:
            `float`: RTD measured temperature (°C)
        """
        self.__enforce_pulse_mode()

        # send command to trigger conversion
        self.start_conversions(ADCNum.ADC_1)

        # send command to read conversion data.
        _, voltage_code, _ = self.__voltage_read(ADCNum.ADC_1)

        # TODO: get RTD calibs from eeprom once added
        return code_to_temperature(
            voltage_code,
            self.r_ref,
            self.rtd_sensor_resistance,
            self.rtd_sensor_resistance_variation
        )

    def reset(self):
        """
        Reset ADC register values to EdgePi ADC power-on state.
        Note this state differs from ADS1263 default power-on, due to
        application of custom power-on configurations required by EdgePi.
        """
        self.transfer(self.adc_ops.reset_adc())
        self.__reapply_config()

    def __is_data_ready(self):
        # pylint: disable=unused-private-member
        # required for integration testing in test_conversion_times.py
        """Utility for testing conversion times, returns True if ADC indicates new voltage data"""
        read_data = self.transfer([ADCComs.COM_RDATA1.value] + [255] * 6)
        return read_data[1] & 0b01000000

    def __read_registers_to_map(self):
        """
        Reads value of all ADC registers to a dictionary

        Returns:
            `dict`: contains {register addx (int): register_value (int)} pairs
        """
        # get register values
        reg_values = self.__read_register(ADCReg.REG_ID, ADC_NUM_REGS)
        addx = ADCReg.REG_ID.value

        # build dict with (register addx : register_value) pairs.
        reg_dict = {addx + i: reg_values[i] for i in range(ADC_NUM_REGS)}
        _logger.debug(f"__read_registers_to_map: {reg_dict}")

        return reg_dict

    def __is_rtd_on(self):
        """
        Get state of RTD_EN pin (on/off), to use for deciding which
        ADC channels are available for reading.

        Returns:
            `bool`: True if RTD_EN pin is on, False otherwise
        """
        # TODO: this should use self.get_state() instead
        _logger.debug("Checking RTD status")
        idac_reg = self.__get_register_map()
        idac_mag = pack("uint:8", idac_reg.get(ADCReg.REG_IDACMAG.value))
        idac_1 = idac_mag[4:].uint
        status = idac_1 != 0x0
        _logger.debug(f"RTD enabled: {status}")
        return status

    # by default set mux_n's to AINCOM. For diff_mode and rtd_mode, pass in custom mapping.
    # Case 1: user sets to adc_x read ch_x -> mux_n auto mapped to AINCOM
    # Case 2: user sets RTD mode on or off -> mux_n passed in as arg
    # Case 3: user sets any differential mode -> mux_n pass in as arg
    def __get_channel_assign_opcodes(
        self,
        adc_1_mux_p: CH = None,
        adc_2_mux_p: CH = None,
        adc_1_mux_n: CH = CH.AINCOM,
        adc_2_mux_n: CH = CH.AINCOM,
        override_rtd_validation: bool = False
    ):
        """
        Generates OpCodes for assigning positive and negative multiplexers
        of either ADC1 or ADC2 to an ADC input channel. This is needed to allow
        users to assign multiplexers to both ADCs using the same input channel
        number enums.

        Args:
            adc_1_mux_p: input channel to assign to MUXP of ADC1
            adc_2_mux_p: input channel to assign to MUXP of ADC2
            adc_1_mux_n: input channel to assign to MUXN of ADC1
            adc_2_mux_n: input channel to assign to MUXN of ADC2

        Returns:
            `list`: if not empty, contains OpCode(s) for updating multiplexer
                channel assignment for ADC1, ADC2, or both.
        """
        # only update mux_n if mux_p is updated
        if adc_1_mux_p is None:
            adc_1_mux_n = None

        if adc_2_mux_p is None:
            adc_2_mux_n = None

        # no multiplexer config to update
        # TODO: refactor filter_dict to take list arg
        args = filter_dict(locals(), "self", None)
        args = filter_dict(args, "override_rtd_validation")
        if not args:
            return []

        # allowed channels depend on RTD_EN status
        if not override_rtd_validation:
            channels = list(args.values())
            rtd_enabled = self.__is_rtd_on()
            validate_channels_allowed(channels, rtd_enabled)

        adc_mux_updates = {
            ADCReg.REG_INPMUX: (adc_1_mux_p, adc_1_mux_n),
            ADCReg.REG_ADC2MUX: (adc_2_mux_p, adc_2_mux_n),
        }

        opcodes = generate_mux_opcodes(adc_mux_updates)

        return opcodes

    def select_differential(self, adc: ADCNum, diff_mode: DiffMode):
        """
        Select a differential voltage sampling mode for either ADC1 or ADC2

        Args:
            `adc` (ADCNum): identity of ADC to configure

            `diff_mode` (DiffMode): a pair of input channels to sample, or off

        Raises:
            `KeyError`: if adc is not a valid ADCNum enum
        """
        mux_p = diff_mode.value.mux_p
        mux_n = diff_mode.value.mux_n
        mux_properties = {
            ADCNum.ADC_1: {
                "adc_1_analog_in": mux_p,
                "adc_1_mux_n": mux_n,
            },
            ADCNum.ADC_2: {
                "adc_2_analog_in": mux_p,
                "adc_2_mux_n": mux_n,
            },
        }
        diff_update = mux_properties[adc]
        self.__config(**diff_update)

    def __validate_no_rtd_conflict(self, updates: dict):
        """
        Checks no RTD related properties are being updated if RTD mode is enabled

        Args:
            `updates` (dict): updates formatted as {'property_name': update_value}

        Raises:
            `RTDEnabledError`: if RTD is enabled and an RTD related property is in updates
        """
        # ADC2 channel setting conflicts with RTD handled during channel mapping
        if not self.__is_rtd_on():
            return

        rtd_properties = RTDModes.RTD_ON.value.keys()
        for update in updates:
            if update in rtd_properties:
                raise RTDEnabledError(
                    f"ADC property '{update}' cannot be updated while RTD is enabled"
                )

    def rtd_mode(self, enable: bool):
        """
        Enable or disable RTD. Note, this will reconfigure ADC2 to read FLOAT input channel
        if ADC2 is configured to read RTD pins AIN4-AIN7. To read RTD values aftering enabling
        RTD mode here, call either `read_voltage` or `single_sample`, depending on which
        conversion mode ADC1 is currently configured to.

        Args:
            `enable` (bool): True to enable RTD, False to disable
        """
        # TODO: enable RTD_EN switching pin
        if enable:
            # check if adc_2 is reading RTD pins, remap channels if needed
            mux_reg = self.__get_register_map()
            adc2_mux = pack("uint:8", mux_reg.get(ADCReg.REG_ADC2MUX.value))
            muxs = [adc2_mux[:4].uint, adc2_mux[4:].uint]
            if any(mux not in [x.value for x in AllowedChannels.RTD_ON.value] for mux in muxs):
                updates = RTDModes.RTD_ON.value | {
                    "adc_2_analog_in": CH.FLOAT,
                    "adc_2_mux_n": CH.AINCOM,
                }
            else:
                updates = RTDModes.RTD_ON.value
        else:
            updates = RTDModes.RTD_OFF.value

        self.__config(**updates, override_rtd_validation=True)

    @staticmethod
    def __extract_mux_args(args: dict) -> dict:
        """
        Checks args dictionary for input multiplexer settings

        Args:
            `args` (dict): args formatted as {'arg_name': value}

        Returns:
            `dict`: input multiplexer args formatted as {'mux_name': value}
        """
        mux_arg_names = {
            "adc_1_analog_in": "adc_1_mux_p",
            "adc_2_analog_in": "adc_2_mux_p",
            "adc_1_mux_n": "adc_1_mux_n",
            "adc_2_mux_n": "adc_2_mux_n",
        }
        only_mux_args = {}
        for arg, mux_name in mux_arg_names.items():
            if args.get(arg) is not None:
                only_mux_args[mux_name] = args[arg]
        return only_mux_args

    def __config(
        self,
        adc_1_analog_in: CH = None,
        adc_2_analog_in: CH = None,
        adc_1_mux_n: CH = None,
        adc_2_mux_n: CH = None,
        adc_1_data_rate: ADC1DataRate = None,
        adc_2_data_rate: ADC2DataRate = None,
        filter_mode: FilterMode = None,
        conversion_mode: ConvMode = None,
        checksum_mode: CheckMode = None,
        reset_clear: ADCPower = None,
        idac_1_mux: IDACMUX = None,
        idac_2_mux: IDACMUX = None,
        idac_1_mag: IDACMAG = None,
        idac_2_mag: IDACMAG = None,
        pos_ref_inp: REFMUX = None,
        neg_ref_inp: REFMUX = None,
        override_updates_validation: bool = False,
        override_rtd_validation: bool = False,
    ):
        """
        Configure all ADC settings, either collectively or individually.
        Warning: for internal EdgePi developer use only, users should use
        `set_config()` for modifying settings instead.

        Args:
            `adc_1_analog_in` (ADCChannel): input voltage channel to map to ADC1 mux_p
            `adc_2_analog_in` (ADCChannel): input voltage channel to map to ADC2 mux_p
            `adc_1_mux_n` (ADCChannel): input voltage channel to map to ADC1 mux_n
            `adc_2_mux_n` (ADCChannel): input voltage channel to map to ADC1 mux_n
            `adc_1_data_rate` (ADC1DataRate): ADC1 data rate in samples per second
            `adc_2_data_rate` (ADC2DataRate): ADC2 data rate in samples per second,
            `filter_mode` (FilterMode): filter mode for both ADC1 and ADC2.
                Note this affects data rate. Please refer to module documentation
                for more information.
            `conversion_mode` (ConvMode): set conversion mode for ADC1.
                Note, ADC2 runs only in continuous conversion mode.
            `checksum_mode` (CheckMode): set mode for CHECK byte
            `reset_clear` (ADCPower): set state of ADC RESET bit
            `idac_1_mux` (IDACMUX): set analog input pin to connect IDAC1
            `idac_2_mux` (IDACMUX): set analog input pin to connect IDAC2
            `idac_1_mag` (IDACMAG): set the current value for IDAC1
            `idac_2_mag` (IDACMAG): set the current value for IDAC2
            `pos_ref_inp` (REFMUX): set the positive reference input
            `neg_ref_inp` (REFMUX): set the negative reference input
            `override_updates_validation` (bool): set to True to override post-update validation
            `override_rtd_validation` (bool): turn off RTD property validation for RTD mode updates
        """
        # pylint: disable=unused-argument

        # filter out self and None args
        args = filter_dict(locals(), "self", None)
        _logger.debug(f"__config: args after filtering out None defaults:\n\n{args}\n\n")

        # permit updates by rtd_mode() to turn RTD off when it's on, validate other updates
        if not override_rtd_validation:
            self.__validate_no_rtd_conflict(args)

        # get opcodes for mapping multiplexers
        mux_args = self.__extract_mux_args(args)
        ops_list = self.__get_channel_assign_opcodes(
            **mux_args, override_rtd_validation=override_rtd_validation
        )

        # extract OpCode type args, since args may contain non-OpCode args
        args = list(args.values())
        ops_list += [
            entry.value
            for entry in args
            if issubclass(entry.__class__, Enum) and isinstance(entry.value, OpCode)
        ]

        # get current register values
        reg_values = self.__get_register_map()
        _logger.debug(f"__config: register values before updates:\n\n{reg_values}\n\n")

        # get codes to update register values
        _logger.debug(f"opcodes = {ops_list}")
        updated_reg_values = apply_opcodes(dict(reg_values), ops_list)
        _logger.debug(f"__config: register values after updates:\n\n{reg_values}\n\n")

        # write updated reg values to ADC using a single write.
        data = [entry["value"] for entry in updated_reg_values.values()]
        self.__write_register(ADCReg.REG_ID, data)

        # update ADC state (for state caching)
        self.__update_cache_map(updated_reg_values)

        # validate updates were applied correctly
        if not override_updates_validation:
            self.__validate_updates(updated_reg_values)

        return updated_reg_values

    def __validate_updates(self, updated_reg_values: dict):
        """
        Validates updated config values have been applied to ADC registers.

        Args:
            updated_reg_values (dict): register values that were applied in latest __config()
        """
        reg_values = self.__get_register_map()

        # check updated value were applied
        for addx, value in reg_values.items():
            observed_val = updated_reg_values[addx]["value"]
            if int(value) != int(updated_reg_values[addx]["value"]):
                _logger.error("__config: failed to update register")
                raise ADCRegisterUpdateError(
                    (
                        "Register failed to update: "
                        f"addx={hex(addx)}, expected: {hex(value)}, observed: {hex(observed_val)}"
                    )
                )

        return True

    def set_config(
        self,
        adc_1_analog_in: CH = None,
        adc_1_data_rate: ADC1DataRate = None,
        adc_2_analog_in: CH = None,
        adc_2_data_rate: ADC2DataRate = None,
        filter_mode: FilterMode = None,
        conversion_mode: ConvMode = None,
        override_updates_validation: bool = False,
    ):
        """
        Configure user accessible ADC settings, either collectively or individually.

        Args:
            `adc_1_analog_in` (ADCChannel): the input voltage channel to measure via ADC1
            `adc_1_data_rate` (ADC1DataRate): ADC1 data rate in samples per second
            `adc_2_analog_in` (ADCChannel): the input voltage channel to measure via ADC2
            `adc_2_data_rate` (ADC2DataRate): ADC2 data rate in samples per second
            `filter_mode` (FilterMode): filter mode for ADC1.
                Note this affects data rate. Please refer to module documentation
                for more information.
            `conversion_mode` (ConvMode): set conversion mode for ADC1.
            `override_updates_validation` (bool): set to True to skip update validation
        """
        # pylint: disable=unused-argument

        args = filter_dict(locals(), "self", None)
        self.__config(**args)

    def get_state(self, override_cache: bool = False) -> ADCState:
        """
        Read the current hardware state of configurable ADC properties

        Args:
            `override_cache` (bool): force SPI read to get hardware state

        Returns:
            ADCState: information about the current ADC hardware state
        """
        reg_values = self.__get_register_map(override_cache)
        return ADCState(reg_values)



""" Helper module for reading thermocouple fault status

    Classes:

        FaultType
        FaultMsg
        Fault

    Functions:
        map_fault_status(Bits)
"""


from dataclasses import dataclass
from enum import Enum, unique
from bitstring import Bits


@unique
class FaultType(Enum):
    """Fault Register bit numbers for each fault type
    Note: the bit numbers are written in reverse order here in order to
    be compatible with the bitstring module, since it considers the MSB
    of a bitstring to be index 0. So CJRANGE actually maps to bit 7 on
    the MAX31856.
    """

    OPEN = 7
    OVUV = 6
    TCLOW = 5
    TCHIGH = 4
    CJLOW = 3
    CJHIGH = 2
    TCRANGE = 1
    CJRANGE = 0


@unique
class FaultMsg(Enum):
    """Debugging messages used in Faults returned to the user"""

    CJRANGE_OK_MSG = "Cold-junction temperature is within normal operating range"
    CJRANGE_BAD_MSG = "Cold-junction temperature is outside normal operating range"
    TCRANGE_OK_MSG = "Thermocouple hot-junction temperature is within normal operating range"
    TCRANGE_BAD_MSG = "Thermocouple hot-junction temperature is outside normal operating range"
    CJHIGH_OK_MSG = (
        "Cold-junction temperature is less than or equal to cold-junction high threshold"
    )
    CJHIGH_BAD_MSG = "Cold-junction temperature is greater than cold-junction high threshold"
    CJLOW_OK_MSG = "Cold-junction temperature greater than or equal to cold-junction low threshold"
    CJLOW_BAD_MSG = "Cold-junction temperature less than cold-junction low threshold"
    TCHIGH_OK_MSG = "Thermocouple temperature is less than or equal to thermocouple high threshold"
    TCHIGH_BAD_MSG = "Thermocouple temperature is greater than thermocouple high threshold"
    TCLOW_OK_MSG = "Thermocouple temperature is greater than or equal to thermocouple low threshold"
    TCLOW_BAD_MSG = "Thermocouple temperature is less than thermocouple low threshold"
    OVUV_OK_MSG = "Input voltage is positive and less than V_DD"
    OVUV_BAD_MSG = "Input voltage is negative or greater than V_DD"
    OPEN_OK_MSG = "No open-circuit or broken thermocouple wires detected"
    OPEN_BAD_MSG = "Open-circuit detected"


@dataclass
class Fault:
    """Represents a Fault Status Register fault

    Attributes:
        name (str): the fault's official MAX31856 name

        fault_type (FaultType): the fault's FaultType type

        err_msg (FaultMsg): a message containing the current fault output

        at_fault (bool): set to True if the fault is currently occurring

        is_masked (bool): set to True if this fault is being masked in MASK register.
                          Note this means the FAULT pin will not assert even if this
                          fault is occurring. Faults are masked by default.
    """

    fault_type: FaultType = None
    err_msg: FaultMsg = None
    at_fault: bool = False
    is_masked: bool = True

    def __repr__(self) -> str:
        msg = (
            "\n\t{"
            + f"""
            Fault Type: {self.fault_type},
            At Fault: {self.at_fault},
            Fault Message: {self.err_msg.value},
            Fault Masked: {self.is_masked},
        """
            + "}\n"
        )
        return msg


_fault_msg_map = {
    FaultType.CJRANGE: (FaultMsg.CJRANGE_OK_MSG, FaultMsg.CJRANGE_BAD_MSG),
    FaultType.TCRANGE: (FaultMsg.TCRANGE_OK_MSG, FaultMsg.TCRANGE_BAD_MSG),
    FaultType.CJHIGH: (FaultMsg.CJHIGH_OK_MSG, FaultMsg.CJHIGH_BAD_MSG),
    FaultType.CJLOW: (FaultMsg.CJLOW_OK_MSG, FaultMsg.CJLOW_BAD_MSG),
    FaultType.TCHIGH: (FaultMsg.TCHIGH_OK_MSG, FaultMsg.TCHIGH_BAD_MSG),
    FaultType.TCLOW: (FaultMsg.TCLOW_OK_MSG, FaultMsg.TCLOW_BAD_MSG),
    FaultType.OVUV: (FaultMsg.OVUV_OK_MSG, FaultMsg.OVUV_BAD_MSG),
    FaultType.OPEN: (FaultMsg.OPEN_OK_MSG, FaultMsg.OPEN_BAD_MSG),
}


def map_fault_status(fault_bits: Bits, fault_masks: Bits) -> dict:
    """Generates a dictionary of Fault objects

    Args:
        fault_bits (bitstring.Bits): Fault Status register bits

        fault_masks (bitstring.Bits): Fault Mask register bits

    Returns:
        a dict containing information on the current status of each Fault Status register bit
    """
    faults_dict = {}

    # check each bit in fault status and fault mask registers to generate
    # Faults
    for tcfault_type in FaultType:
        fault = Fault(fault_type=tcfault_type)

        # get value of ith bit in fault_bits register, either 0 or 1
        fault_bit_value = fault_bits[tcfault_type.value]

        # get message depending on whether ith bit of fault_bits is set or not
        fault.err_msg = _fault_msg_map[tcfault_type][fault_bit_value]

        fault.at_fault = fault_bit_value

        # check ith bit in fault_masks register to see if this fault type is
        # being masked.
        fault.is_masked = fault_masks[tcfault_type.value]

        faults_dict[tcfault_type] = fault

    return faults_dict



"""EdgePi LED array"""


import logging


from edgepi.gpio.gpio_configs import LEDPins
from edgepi.gpio.edgepi_gpio import EdgePiGPIO


class InvalidLEDNameError(Exception):
    """
    Raised if a user attempts to pass an invalid LED name to
    an EdgePiLED method."""


class EdgePiLED:
    """Interact with the EdgePi LED Array"""

    def __init__(self):
        self.gpio_ops = EdgePiGPIO()
        self.log = logging.getLogger(__name__)

    @staticmethod
    def __validate_led_name(led_name: LEDPins):
        """Checks if an LED name is valid"""
        if not isinstance(led_name, LEDPins):
            raise InvalidLEDNameError(
                f"{led_name} is not a valid EdgePi LED name, use LEDPins enums for names."
            )

    def turn_led_on(self, led_name: LEDPins):
        """
        Turn an LED on

        Args:
            `led_name` (LEDPins): name of LED to turn on
        """
        self.__validate_led_name(led_name)
        self.gpio_ops.set_pin_state(led_name.value)
        self.log.info(f"LED with name {led_name.value} has been turned on")

    def turn_led_off(self, led_name: LEDPins):
        """
        Turn an LED off

        Args:
            `led_name` (LEDPins): name of LED to turn off
        """
        self.__validate_led_name(led_name)
        self.gpio_ops.clear_pin_state(led_name.value)
        self.log.info(f"LED with name {led_name.value} has been turned off")

    def toggle_led(self, led_name: LEDPins):
        """
        Toggle an LED to its opposite state

        Args:
            `led_name` (LEDPins): name of LED to toggle
        """
        self.__validate_led_name(led_name)
        self.gpio_ops.toggle_pin(led_name.value)
        self.log.info(f"LED with name {led_name.value} has been toggled to opposite state")

    def get_led_state(self, led_name: LEDPins) -> bool:
        """
        Read an LED's current state (on/off)

        Args:
            `led_name` (LEDPins): name of LED whose state is to be read

        Returns:
            `bool`: True if LED is on, False if LED is off
        """
        self.__validate_led_name(led_name)
        state = self.gpio_ops.read_pin_state(led_name.value)
        msg = "ON" if state else "OFF"
        self.log.info(f"LED with name {led_name.value} is currently {msg}")
        return state



"""unit tests for reg_helper.py module"""


from copy import deepcopy
import pytest
from edgepi.reg_helper.reg_helper import (
    _apply_opcode,
    _add_change_flags,
    apply_opcodes,
    _convert_values_to_dict,
    convert_dict_to_values,
    is_bit_set,
)
from edgepi.tc.tc_constants import (
    AvgMode,
    CJMode,
    ConvMode,
    FaultMode,
    NoiseFilterMode,
    OpenCircuitMode,
    TCAddresses,
    TCOps,
    TCType,
    VoltageMode,
)


@pytest.mark.parametrize(
    "reg_value, opcode, updated_reg_value",
    [
        (0b00000000, ConvMode.AUTO.value, 0b10000000),  # sets bit 7, doesn't set others
        (0b10000000, ConvMode.AUTO.value, 0b10000000),  # doesn't toggle bit 7
        (
            0b01111111,
            ConvMode.AUTO.value,
            0b11111111,
        ),  # sets bit 7, doesn't clear others
        (
            0b11111111,
            ConvMode.SINGLE.value,
            0b01111111,
        ),  # clears bit 7, doesn't clear others
        (0b00000000, ConvMode.SINGLE.value, 0b00000000),  # doesn't toggle bit 7
        (
            0b10000000,
            ConvMode.SINGLE.value,
            0b00000000,
        ),  # clears bit 7, doesn't set others
        (
            0b00000000,
            TCOps.SINGLE_SHOT.value,
            0b01000000,
        ),  # sets bit 6, doesn't set others
        (0b01000000, TCOps.SINGLE_SHOT.value, 0b01000000),  # doesn't toggle bit 6
        (
            0b10111111,
            TCOps.SINGLE_SHOT.value,
            0b11111111,
        ),  # sets bit 6, doesn't clear others
        (
            0b00001000,
            CJMode.ENABLE.value,
            0b00000000,
        ),  # clears bit 3, doesn't set others
        (0b00000000, CJMode.ENABLE.value, 0b00000000),  # doesn't toggle bit 3
        (
            0b11111111,
            CJMode.ENABLE.value,
            0b11110111,
        ),  # clears bit 3, doesn't clear others
        (
            0b00000000,
            CJMode.DISABLE.value,
            0b00001000,
        ),  # sets bit 3, doesn't set others
        (0b00001000, CJMode.DISABLE.value, 0b00001000),  # doesn't toggle bit 3
        (
            0b11111111,
            CJMode.DISABLE.value,
            0b11111111,
        ),  # sets bit 3, doesn't clear others
        (
            0b00000100,
            FaultMode.COMPARATOR.value,
            0b00000000,
        ),  # clears bit 2, doesn't set others
        (0b00000000, FaultMode.COMPARATOR.value, 0b00000000),  # doesn't toggle bit 2
        (
            0b11111111,
            FaultMode.COMPARATOR.value,
            0b11111011,
        ),  # clears bit 2, doesn't clear others
        (
            0b00000000,
            FaultMode.INTERRUPT.value,
            0b00000100,
        ),  # sets bit 2, doesn't set others
        (0b00000100, FaultMode.INTERRUPT.value, 0b00000100),  # doesn't toggle bit 2
        (
            0b11111011,
            FaultMode.INTERRUPT.value,
            0b11111111,
        ),  # sets bit 2, doesn't clear others
        (
            0b00000000,
            TCOps.CLEAR_FAULTS.value,
            0b00000010,
        ),  # sets bit 1, doesn't set others
        (0b00000010, TCOps.CLEAR_FAULTS.value, 0b00000010),  # doesn't toggle bit 1
        (
            0b11111101,
            TCOps.CLEAR_FAULTS.value,
            0b11111111,
        ),  # sets bit 1, doesn't clear others
        (
            0b00000001,
            NoiseFilterMode.HZ_60.value,
            0b00000000,
        ),  # clears bit 0, doesn't set others
        (0b00000000, NoiseFilterMode.HZ_60.value, 0b00000000),  # doesn't toggle bit 0
        (
            0b11111111,
            NoiseFilterMode.HZ_60.value,
            0b11111110,
        ),  # clears bit 0, doesn't clear others
        (
            0b00000000,
            NoiseFilterMode.HZ_50.value,
            0b00000001,
        ),  # sets bit 0, doesn't set others
        (0b00000001, NoiseFilterMode.HZ_50.value, 0b00000001),  # doesn't toggle bit 0
        (
            0b11111110,
            NoiseFilterMode.HZ_50.value,
            0b11111111,
        ),  # sets bit 0, doesn't clear others
        (
            0b11110000,
            AvgMode.AVG_1.value,
            0b00000000,
        ),  # clears high byte, doesn't set low byte
        (0b00000000, AvgMode.AVG_1.value, 0b00000000),  # doesn't toggle high byte
        (
            0b11111111,
            AvgMode.AVG_1.value,
            0b00001111,
        ),  # clears high byte, doesn't clear low byte
        (
            0b00000000,
            AvgMode.AVG_2.value,
            0b00010000,
        ),  # only sets intended high byte bits
        (
            0b00010000,
            AvgMode.AVG_2.value,
            0b00010000,
        ),  # doesn't toggle intended high byte bits
        (
            0b11101111,
            AvgMode.AVG_2.value,
            0b00011111,
        ),  # only clears intended high byte bits
        (0b00000000, AvgMode.AVG_4.value, 0b00100000),  # only sets high byte bits
        (
            0b00100000,
            AvgMode.AVG_4.value,
            0b00100000,
        ),  # doesn't toggle intended high byte bits
        (
            0b11011111,
            AvgMode.AVG_4.value,
            0b00101111,
        ),  # only clears intended high byte bits
        (0b00000000, AvgMode.AVG_8.value, 0b00110000),  # only sets high byte bits
        (0b00110000, AvgMode.AVG_8.value, 0b00110000),  # doesn't toggle high byte bits
        (0b11001111, AvgMode.AVG_8.value, 0b00111111),  # only clears high byte bits
        (0b00000000, AvgMode.AVG_16.value, 0b01000000),  # only sets high byte bits
        (0b01000000, AvgMode.AVG_16.value, 0b01000000),  # doesn't toggle high byte bits
        (0b10111111, AvgMode.AVG_16.value, 0b01001111),  # only clears high byte bits
        (
            0b00001111,
            TCType.TYPE_B.value,
            0b00000000,
        ),  # clears low byte, doesn't set high byte
        (0b00000000, TCType.TYPE_B.value, 0b00000000),  # doesn't toggle low byte
        (
            0b11111111,
            TCType.TYPE_B.value,
            0b11110000,
        ),  # clears low byte, doesn't clear high byte
        (0b00000000, TCType.TYPE_E.value, 0b00000001),  # only sets low byte bits
        (0b00000001, TCType.TYPE_E.value, 0b00000001),  # doesn't toggle low byte
        (0b11111110, TCType.TYPE_E.value, 0b11110001),  # only clears low byte bits
        (0b00000000, TCType.TYPE_J.value, 0b00000010),  # only sets low byte bits
        (0b00000010, TCType.TYPE_J.value, 0b00000010),  # doesn't toggle low byte
        (0b11111101, TCType.TYPE_J.value, 0b11110010),  # only clears low byte bits
        (0b00000000, TCType.TYPE_K.value, 0b00000011),  # only sets low byte bits
        (0b00000011, TCType.TYPE_K.value, 0b00000011),  # doesn't toggle low byte
        (0b11111100, TCType.TYPE_K.value, 0b11110011),  # only clears low byte bits
        (0b00000000, TCType.TYPE_N.value, 0b00000100),  # only sets low byte bits
        (0b00000100, TCType.TYPE_N.value, 0b00000100),  # doesn't toggle low byte
        (0b11111011, TCType.TYPE_N.value, 0b11110100),  # only clears low byte bits
        (0b00000000, TCType.TYPE_R.value, 0b00000101),  # only sets low byte bits
        (0b00000101, TCType.TYPE_R.value, 0b00000101),  # doesn't toggle low byte
        (0b11111010, TCType.TYPE_R.value, 0b11110101),  # only clears low byte bits
        (0b00000000, TCType.TYPE_S.value, 0b00000110),  # only sets low byte bits
        (0b00000110, TCType.TYPE_S.value, 0b00000110),  # doesn't toggle low byte
        (0b11111001, TCType.TYPE_S.value, 0b11110110),  # only clears low byte bits
        (0b00000000, TCType.TYPE_T.value, 0b00000111),  # only sets low byte bits
        (0b00000111, TCType.TYPE_T.value, 0b00000111),  # doesn't toggle low byte
        (0b11111000, TCType.TYPE_T.value, 0b11110111),  # only clears low byte bits
        (0b00000000, VoltageMode.GAIN_8.value, 0b00001000),  # only sets low byte bits
        (0b00001000, VoltageMode.GAIN_8.value, 0b00001000),  # doesn't toggle low byte
        (0b11110111, VoltageMode.GAIN_8.value, 0b11111000),  # only clears low byte bits
        (0b00000000, VoltageMode.GAIN_32.value, 0b00001100),  # only sets low byte bits
        (0b00001100, VoltageMode.GAIN_32.value, 0b00001100),  # doesn't toggle low byte
        (
            0b11110011,
            VoltageMode.GAIN_32.value,
            0b11111100,
        ),  # only clears low byte bits
        (0b11111111, OpenCircuitMode.DISABLED.value, 0b11001111),  # only clears oc bits
        (
            0b00000000,
            OpenCircuitMode.DISABLED.value,
            0b00000000,
        ),  # doesn't toggle oc bits or set other bits
        (
            0b00000000,
            OpenCircuitMode.LOW_INPUT_IMPEDANCE.value,
            0b00010000,
        ),  # only sets oc bits
        (
            0b00010000,
            OpenCircuitMode.LOW_INPUT_IMPEDANCE.value,
            0b00010000,
        ),  # doesn't toggle oc bits
        (
            0b11111111,
            OpenCircuitMode.LOW_INPUT_IMPEDANCE.value,
            0b11011111,
        ),  # doesn't clear other bits
        (
            0b00000000,
            OpenCircuitMode.MED_INPUT_IMPEDANCE.value,
            0b00100000,
        ),  # only sets oc bits
        (
            0b00100000,
            OpenCircuitMode.MED_INPUT_IMPEDANCE.value,
            0b00100000,
        ),  # doesn't toggle oc bits
        (
            0b11111111,
            OpenCircuitMode.MED_INPUT_IMPEDANCE.value,
            0b11101111,
        ),  # doesn't clear other bits
        (
            0b00000000,
            OpenCircuitMode.HIGH_INPUT_IMPEDANCE.value,
            0b00110000,
        ),  # only sets oc bits
        (
            0b00110000,
            OpenCircuitMode.HIGH_INPUT_IMPEDANCE.value,
            0b00110000,
        ),  # doesn't toggle oc bits
        (
            0b11111111,
            OpenCircuitMode.HIGH_INPUT_IMPEDANCE.value,
            0b11111111,
        ),  # doesn't clear other bits
    ],
)
def test_apply_opcode(reg_value, opcode, updated_reg_value):
    assert _apply_opcode(reg_value, opcode) == updated_reg_value


@pytest.mark.parametrize(
    "reg_values",
    [
        ({0x0: {"value": 0x255}}),
        ({0x0: {"value": 0x255}, 0x1: {"value": 0x255}, 0x2: {"value": 0x255}}),
        ({0x0: {}, 0x1: {"value": 0x255}, 0x2: {}}),
    ],
)
def test_add_change_flags_adds_flags(reg_values):
    _add_change_flags(reg_values)
    for key in reg_values:
        assert reg_values[key]["is_changed"] is False


@pytest.mark.parametrize(
    "reg_values, opcodes",
    [
        ({}, []),
        ({0x0: {}}, []),
        ({}, [AvgMode.AVG_1]),
    ],
)
def test_apply_opcodes_raises(reg_values, opcodes):
    with pytest.raises(Exception) as err:
        apply_opcodes(reg_values, opcodes)
    assert "register_values and opcodes args must both be non-empty" in str(err.value)


@pytest.mark.parametrize(
    "reg_values, opcodes, out",
    [
        (
            {TCAddresses.CR1_W.value: 0x0},
            [AvgMode.AVG_1.value],
            {
                TCAddresses.CR1_W.value: {
                    "value": AvgMode.AVG_1.value.op_code,
                    "is_changed": True,
                }
            },
        ),
        (
            {TCAddresses.CR0_W.value: 0x0, TCAddresses.CR1_W.value: 0x0},
            [AvgMode.AVG_1.value],
            {
                TCAddresses.CR0_W.value: {"value": 0x0, "is_changed": False},
                TCAddresses.CR1_W.value: {
                    "value": AvgMode.AVG_1.value.op_code,
                    "is_changed": True,
                },
            },
        ),
        (
            {TCAddresses.CR0_W.value: 0x0, TCAddresses.CR1_W.value: 0x0},
            [AvgMode.AVG_1.value, ConvMode.AUTO.value],
            {
                TCAddresses.CR0_W.value: {
                    "value": ConvMode.AUTO.value.op_code,
                    "is_changed": True,
                },
                TCAddresses.CR1_W.value: {
                    "value": AvgMode.AVG_1.value.op_code,
                    "is_changed": True,
                },
            },
        ),
        (
            {TCAddresses.CR0_W.value: 0x0, TCAddresses.CR1_W.value: 0x0},
            [AvgMode.AVG_1.value, ConvMode.AUTO.value],
            {
                TCAddresses.CR0_W.value: {
                    "value": ConvMode.AUTO.value.op_code,
                    "is_changed": True,
                },
                TCAddresses.CR1_W.value: {
                    "value": AvgMode.AVG_1.value.op_code,
                    "is_changed": True,
                },
            },
        ),
    ],
)
def test_apply_opcodes(reg_values, opcodes, out):
    assert apply_opcodes(reg_values, opcodes) == out


@pytest.mark.parametrize(
    "reg_map",
    [
        ({TCAddresses.CR0_R.value: 0xFF}),
        (
            {
                TCAddresses.CR0_R.value: 0xFF,
                TCAddresses.CR1_R.value: 0x00,
                TCAddresses.CJLF_R.value: 0x5F,
            }
        ),
    ],
)
def test_convert_values_to_dict(reg_map):
    map_copy = deepcopy(reg_map)
    _convert_values_to_dict(reg_map)
    for key in reg_map:
        assert isinstance(reg_map[key], dict)  # assert changed format to addx, dict pairs
        assert (
            reg_map[key]["value"] == map_copy[key]
        )  # assert value field in dict, and reg_value copied correctly

@pytest.mark.parametrize('reg_dict, result',[({3:{'value':32,'is_changed':False}}, {3:32}),
                                             ({3:{'value':32,'is_changed':True}}, {3:32}),
                                             ({3:{'value':32,'is_changed':False},
                                               5:{'value':11,'is_changed':True}}, {3:32, 5:11})])
def test_convert_dict_to_values(reg_dict, result):
    reg_dict = convert_dict_to_values(reg_dict)
    assert reg_dict == result


@pytest.mark.parametrize("reg_val, bit_mask, result", [
    (0b00000001, 0b11111110, True),
    (0b00000000, 0b11111110, False),
    (0b00000010, 0b11111101, True),
    (0b00000000, 0b11111101, False),
    (0b00000100, 0b11111011, True),
    (0b00000000, 0b11111011, False),
    (0b00001000, 0b11110111, True),
    (0b00000000, 0b11110111, False),
    (0b00010000, 0b11101111, True),
    (0b00000000, 0b11101111, False),
    (0b00100000, 0b11011111, True),
    (0b00000000, 0b11011111, False),
    (0b01000000, 0b10111111, True),
    (0b00000000, 0b10111111, False),
    (0b10000000, 0b01111111, True),
    (0b00000000, 0b01111111, False),
    (0xFF, 0b11111110, True),
    (0b11111110, 0b11111110, False),
    (0xFF, 0b11111101, True),
    (0b11111101, 0b11111101, False),
    (0xFF, 0b11111011, True),
    (0b11111011, 0b11111011, False),
    (0xFF, 0b11110111, True),
    (0b11110111, 0b11110111, False),
    (0xFF, 0b11101111, True),
    (0b11101111, 0b11101111, False),
    (0xFF, 0b11011111, True),
    (0b11011111, 0b11011111, False),
    (0xFF, 0b10111111, True),
    (0b10111111, 0b10111111, False),
    (0xFF, 0b01111111, True),
    (0b01111111, 0b01111111, False),
])
def test_is_bit_set(reg_val, bit_mask, result):
    assert is_bit_set(reg_val, bit_mask) == result
